# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import time
from enum import Enum
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth.hashers import make_password, check_password
from django.contrib.postgres.fields import JSONField
from django.core.validators import RegexValidator
from django.db import models
from django.db.models.options import Options
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

if TYPE_CHECKING:
    from .people import ResponsibilityPerson

from .. import utils

phone_regex = RegexValidator(regex=r'^\+?1?\d{9,15}$',
                             message=_("Phone number must be entered in the format: "
                                       "'050123123' or '+35850123123'. "
                                       "Up to 15 digits allowed."))
postalcode_regex = RegexValidator(
    regex=r'^\d{5}$',
    message=_('Enter a valid postal code.'),
    code='invalid',
)

class ContactDetailsMixin(models.Model):
    class Meta:
        abstract = True

    def __str__(self):
        if self.forenames or self.surname:
            return '%s %s' % (self.forenames.title(), self.surname.title())
        else:
            return self.email

    @property
    def address(self):
        return ('%s\n%s %s' % (
            self.street_address.title(), self.postal_code, self.city.title())).strip()

    forenames = models.CharField(max_length=64, verbose_name=_('Forename(s)'))
    surname = models.CharField(max_length=64, verbose_name=_('Surname'))
    street_address = models.CharField(max_length=128, blank=False,
                                      verbose_name=_('Street address'))
    postal_code = models.CharField(max_length=32, blank=False,
                                   verbose_name=_('Zip/Postal code'),
                                   validators=[postalcode_regex])
    city = models.CharField(max_length=32, blank=False, verbose_name=_('City'))
    email = models.EmailField(blank=False, verbose_name=_('Email'), db_index=True)
    phone_number = models.CharField(max_length=32, validators=[phone_regex], blank=False,
                                    verbose_name=_('Phone number'))

    @property
    def contact_details(self):
        yield _('Name'), '%s %s' % (self.forenames.title(), self.surname.title())
        if self.email:
            yield _('Email'), self.email
        if self.phone_number:
            yield _('Phone number'), self.phone_number
        if self.address:
            yield _('Address'), '\n' + self.address

    @property
    def contact_display(self):
        return '\n'.join('%s: %s' % (k, v) for k, v in self.contact_details)


class ContactDetailsMixinEmail(ContactDetailsMixin):
    class Meta:
        abstract = True

ContactDetailsMixinEmail._meta.get_field('street_address').blank = True
ContactDetailsMixinEmail._meta.get_field('postal_code').blank = True
ContactDetailsMixinEmail._meta.get_field('city').blank = True
ContactDetailsMixinEmail._meta.get_field('phone_number').blank = True


class NameDescriptionMixin(models.Model):
    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    name = models.CharField(max_length=256, verbose_name=_('Name'))
    description = models.TextField(blank=True, verbose_name=_('Description'))


class CopyMixin:
    _meta: Options
    def create_copy(self):
        fr = self.__class__()
        for field in self._meta.fields:
            setattr(fr, field.name, getattr(self, field.name))
        fr.pk = None
        return fr


class SubitemMixin(CopyMixin):
    subitem_name: str
    _counter: int

    def __init__(self, *args, **kwargs):
        self._responsibles = set()
        super().__init__(*args, **kwargs)

    @cached_property
    def sub_items(self):
        return getattr(self, self.subitem_name + '_set').all()

    def has_responsible(self, r: 'ResponsibilityPerson') -> bool:
        return r in self._responsibles


class PasswordMixin(models.Model):
    """
    New 'password' is generated every time user requests a auth email to be sent
    to him. Password will expire after AUTH_KEY_EXPIRE_DAYS. We will store
    AUTH_STORE_KEYS number of most recent keys in a json storage.
    """

    AUTH_VIEW: str

    class Meta:
        abstract = True

    class PasswordStatus(Enum):
        PASSWORD_EXPIRED = object()
        PASSWORD_OK = True
        PASSWORD_NOK = False

    # New style auth link hash
    auth_keys_hash_storage = JSONField(default=[])  # List of (hash, expire) tuples

    # TODO: remove this field (as well as views using it) when all users are having new auth_key_hash set up.
    secret_key = models.CharField(max_length=36, default=utils.generate_uuid, db_index=True,
                                  unique=True,
                                  verbose_name=_('Secret key'))

    def make_new_password(self) -> str:
        valid_hashes = []
        for key, expire in self.auth_keys_hash_storage:
            if expire > time.time():
                valid_hashes.append((key, expire))

        password = utils.generate_uuid()

        auth_key_hash = make_password(password)
        auth_key_expire: datetime.datetime = (timezone.now() +
                           datetime.timedelta(days=getattr(settings, 'AUTH_KEY_EXPIRE_DAYS', 90)))

        valid_hashes.append((auth_key_hash, auth_key_expire.timestamp()))
        self.auth_keys_hash_storage = valid_hashes[-getattr(settings, 'AUTH_STORE_KEYS', 10):]
        self.save(update_fields=['auth_keys_hash_storage'])
        return password

    def make_new_auth_url(self) -> str:
        url = settings.SERVER_URL + reverse(self.AUTH_VIEW, args=(self.pk,
                                                                  self.make_new_password(),))
        return url

    def check_auth_key(self, password: str) -> PasswordStatus:
        for key, expire_timestamp in reversed(self.auth_keys_hash_storage):
            if check_password(password, key):
                if expire_timestamp < time.time():
                    return self.PasswordStatus.PASSWORD_EXPIRED
                return self.PasswordStatus.PASSWORD_OK

        return self.PasswordStatus.PASSWORD_NOK





# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.
import json
import logging
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.template import Context, Template
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from python_http_client import HTTPError
from raven.contrib.django.raven_compat.models import client

logger = logging.getLogger(__name__)
from .mixins import CopyMixin

if TYPE_CHECKING:
    from .serviceform import ServiceForm

class EmailMessage(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)
    template = models.ForeignKey('serviceform.EmailTemplate', null=True, on_delete=models.SET_NULL)
    from_address = models.CharField(max_length=256)
    to_address = models.CharField(max_length=256)
    subject = models.CharField(max_length=256)
    content = models.TextField()
    sent_at = models.DateTimeField(null=True)
    context = models.TextField(default="{}")  # JSONified context variables

    def __str__(self):
        return '<EmailMessage %s to %s>' % (self.pk, self.to_address)

    @cached_property
    def context_dict(self) -> Context:
        return Context(json.loads(self.context))

    def content_display(self) -> str:
        return Template(self.content).render(self.context_dict)

    content_display.short_description = _('Content')

    def subject_display(self) -> str:
        return Template(self.subject).render(self.context_dict)

    subject_display.short_description = _('Subject')

    def _cleanup_context(self) -> None:
        context = json.loads(self.context)
        if 'url' in context:
            # Remove URL from email message, as it contains password
            context['url'] = 'http://***password*removed***'
            self.context = json.dumps(context)
            self.save(update_fields=['context'])

    def send(self) -> None:
        logger.info('Sending email to %s', self.to_address)
        body = self.content_display()
        html_body = render_to_string('serviceform/email.html', context={'body': body})
        headers = {'List-Unsubscribe': '<%s>' % self.context_dict['list_unsubscribe']}
        mail = EmailMultiAlternatives(subject=self.subject_display(),
                                      body=body,
                                      from_email=settings.SERVER_EMAIL,
                                      headers=headers,
                                      to=[self.to_address])
        mail.attach_alternative(html_body, 'text/html')
        try:
            emails = mail.send()
        except Exception:
            client.captureException()
            logger.exception('Problem in email sending to %s. Will try again later', self.to_address)
            emails = 0

        if emails == 1:
            self.sent_at = timezone.now()
            self.save(update_fields=['sent_at'])
            self._cleanup_context()
        else:
            logger.error('Email message to %s could not be sent', self)

    @classmethod
    def make(cls, template: 'EmailTemplate', context_dict: dict, address: str,
             send: bool=False) -> 'EmailMessage':
        logger.info('Creating email to %s', address)
        msg = cls.objects.create(template=template, to_address=address,
                                 from_address=settings.SERVER_EMAIL,
                                 subject=template.subject, content=template.content,
                                 context=json.dumps(context_dict))
        if send:
            msg.send()
        return msg


class EmailTemplate(CopyMixin, models.Model):
    class Meta:
        verbose_name = _('Email template')
        verbose_name_plural = _('Email templates')

    def __str__(self):
        return self.name

    name = models.CharField(_('Template name'), max_length=256)
    subject = models.CharField(_('Subject'), max_length=256)
    content = models.TextField(_('Content'), help_text=_(
        'Following context may (depending on topic) be available for both subject and content: '
        '{{responsible}}, {{participant}}, {{last_modified}}, {{form}}, {{url}}, {{contact}}'))
    form = models.ForeignKey('serviceform.ServiceForm', on_delete=models.CASCADE)

    @classmethod
    def make(cls, name: str, form: 'ServiceForm', content: str, subject: str):
        return cls.objects.create(name=name, form=form, subject=subject, content=content)


# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.


import time
import logging

from django.core.management import BaseCommand
from django.utils.translation import activate
from django.conf import settings
from serviceform.serviceform.models import EmailMessage
from serviceform.serviceform.utils import DelayedKeyboardInterrupt


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    args = '-'
    help = 'Send emails'

    def handle(self, *args, **kwargs):
        activate(settings.LANGUAGE_CODE)
        while True:
            msgs = EmailMessage.objects.filter(sent_at__isnull=True)
            for m in msgs:
                with DelayedKeyboardInterrupt():
                    m.send()
            time.sleep(5)



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

import logging

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, Http404, HttpRequest, HttpResponse
from django.shortcuts import render, get_object_or_404, redirect
from django.utils.translation import ugettext_lazy as _

from .. import forms, models
from ..utils import clean_session, user_has_serviceform_permission, expire_auth_link, decode
from .decorators import require_authenticated_participant, require_published_form

logger = logging.getLogger(__name__)


@require_authenticated_participant
def contact_details(request: HttpRequest, participant: models.Participant) -> HttpResponse:
    if participant and participant.status == models.Participant.STATUS_FINISHED:
        return HttpResponseRedirect(reverse('submitted'))

    form = forms.ContactForm(instance=participant, user=request.user)

    if request.method == 'POST':
        form = forms.ContactForm(request.POST, instance=participant, user=request.user)
        if form.is_valid():
            participant = form.save()
            if participant.form.is_published:
                return participant.redirect_next(request)
            else:
                participant.status = models.Participant.STATUS_FINISHED
                participant.save(update_fields=['status'])
                return HttpResponseRedirect(reverse('submitted'))

    return render(request, 'serviceform/participation/contact_view.html',
                  {'form': form,
                   'participant': participant,
                   'service_form': participant.form,
                   'bootstrap_checkbox_disabled': True})


@require_authenticated_participant
@require_published_form
def email_verification(request: HttpRequest, participant: models.Participant) -> HttpResponse:
    service_form = participant.form
    if request.session.get('verification_sent', '') != participant.email:
        participant.send_participant_email(models.Participant.EmailIds.EMAIL_VERIFICATION)
        request.session['verification_sent'] = participant.email
    else:
        messages.warning(request,
                         _('Verification email already sent '
                           'to {}, not sending again.').format(participant.email))
    return render(request, 'serviceform/participation/email_verification.html',
                  {'service_form': service_form,
                   'participant': participant,
                   'bootstrap_checkbox_disabled': True})


@require_authenticated_participant
@require_published_form
def participation(request: HttpRequest, participant: models.Participant,
                  cat_num: int) -> HttpResponse:
    cat_num = int(cat_num)
    service_form = participant.form
    service_form.init_counters()

    category = service_form.sub_items[cat_num] if service_form.flow_by_categories else None
    num_categories = len(service_form.sub_items) if category else 0

    if participant.can_access_view(
            participant.next_view_name) or service_form.allow_skipping_categories:
        max_cat = num_categories
    else:
        max_cat = int(request.session.get('max_category', 0))

    if cat_num > max_cat:
        return HttpResponseRedirect(reverse('participation', args=(max_cat,)))

    form = forms.ParticipationForm(request, participant, category)
    if request.method == 'POST':
        form = forms.ParticipationForm(request, participant, category, request.POST)
        if form.is_valid():
            form.save()
            cat_num += 1
            request.session['max_category'] = max(cat_num, max_cat)
            if cat_num >= num_categories:
                return participant.redirect_next(request)
            else:
                return HttpResponseRedirect(reverse('participation', args=(cat_num,)))

    return render(request, 'serviceform/participation/participation_view.html',
                  {'form': form,
                   'service_form': service_form,
                   'cat_num': cat_num,
                   'max_cat': max_cat})


@require_authenticated_participant
@require_published_form
def questions(request: HttpRequest, participant: models.Participant) -> HttpResponse:
    if not participant.form.questions:
        return participant.redirect_next(request)

    form = forms.QuestionForm(request, participant)

    if request.method == 'POST':
        form = forms.QuestionForm(request, participant, request.POST)
        if form.is_valid():
            form.save()
            return participant.redirect_next(request)

    return render(request, 'serviceform/participation/question_view.html',
                  {'form': form, 'service_form': participant.form})


@require_authenticated_participant
@require_published_form
def preview(request: HttpRequest, participant: models.Participant) -> HttpResponse:
    if request.method == 'POST' and 'submit' in request.POST:
        return participant.redirect_next(request, message=False)
    else:
        return render(request, 'serviceform/participation/preview_view.html',
                      {'service_form': participant.form, 'participant': participant})


@require_authenticated_participant
def submitted(request: HttpRequest, participant: models.Participant) -> HttpResponse:
    participant.finish()
    clean_session(request)
    return render(request, 'serviceform/participation/submitted_view.html',
                  {'service_form': participant.form, 'participant': participant})


@require_authenticated_participant(check_flow=False)
def send_auth_link(request: HttpRequest, participant: models.Participant,
                   email: str) -> HttpResponse:
    if not email:
        raise Http404
    p = get_object_or_404(models.Participant, email=email, form_revision__form=participant.form)
    p.send_participant_email(p.EmailIds.RESEND)
    messages.add_message(request, messages.INFO,
                         _('Authentication link was sent to email address {}.').format(email))
    return HttpResponseRedirect(reverse('contact_details'))


def auth_participant_common(request: HttpRequest, participant: models.Participant, next_view: str,
                            email_verified: bool=True) -> HttpResponse:
    if not participant.email_verified and email_verified:
        participant.email_verified = True
        messages.info(request,
                      _('Your email {} is now verified successfully!').format(participant.email))

    if participant.status == models.Participant.STATUS_FINISHED:
        participant.status = models.Participant.STATUS_UPDATING
    elif participant.status == models.Participant.STATUS_INVITED:
        participant.status = models.Participant.STATUS_ONGOING
    if participant.form_revision != participant.form_revision.form.current_revision:
        participant.last_finished_view = ''
    participant.form_revision = participant.form_revision.form.current_revision
    participant.save(
        update_fields=['status', 'form_revision', 'last_finished_view', 'email_verified'])
    request.session['authenticated_participant'] = participant.pk
    return redirect(next_view)


def authenticate_participant_old(request: HttpRequest, uuid: str,
                                 next_view: str='contact_details') -> HttpResponse:
    """
    Old insecure authentication of participant. Just expire link and send new authentication url.
    """
    if not uuid:
        raise Http404
    clean_session(request)
    participant = get_object_or_404(models.Participant.objects.all(), secret_key=uuid)
    return expire_auth_link(request, participant)


def authenticate_participant(request: HttpRequest, participant_id: int, password: str,
                             next_view: str='contact_details') -> HttpResponse:
    clean_session(request)
    participant = get_object_or_404(models.Participant.objects.all(), pk=participant_id)
    result = participant.check_auth_key(password)
    if result == participant.PasswordStatus.PASSWORD_NOK:
        messages.error(request, _(
            "Given URL might be expired. Please give your "
            "email address and we'll send you a new link"))
        return redirect('send_participant_email', participant.form.slug)

    elif result == participant.PasswordStatus.PASSWORD_EXPIRED:
        return expire_auth_link(request, participant)

    return auth_participant_common(request, participant, next_view)


@login_required(login_url=settings.LOGIN_URL)
def authenticate_participant_mock(request: HttpRequest, participant_id: int,
                                  next_view: str='contact_details') -> HttpResponse:
    clean_session(request)
    participant = get_object_or_404(models.Participant.objects.all(), pk=participant_id)
    user_has_serviceform_permission(request.user, participant.form, raise_permissiondenied=True)
    return auth_participant_common(request, participant, next_view, email_verified=False)


@require_authenticated_participant(check_flow=False)
def delete_participation(request: HttpRequest, participant: models.Participant) -> HttpResponse:
    form = forms.DeleteParticipationForm()
    service_form = participant.form
    if request.method == 'POST':
        form = forms.DeleteParticipationForm(request.POST)
        if form.is_valid():
            logger.info('Deleting participant %s, per request.', participant)
            participant.delete()
            clean_session(request)
            messages.info(request, _('Your participation was deleted'))
            return redirect('password_login', service_form.slug)

    return render(request, 'serviceform/participation/delete_participation.html',
                  {'form': form, 'service_form': service_form,
                   'bootstrap_checkbox_disabled': True})


def verify_email(request: HttpRequest, participant_id: int, password: str) -> HttpResponse:
    return authenticate_participant(request, participant_id, password, 'participation')


def unsubscribe(request: HttpRequest, secret_id: str) -> HttpResponse:
    participant = get_object_or_404(models.Participant.objects, pk=decode(secret_id))
    participant.send_email_allowed = False
    participant.save(update_fields=['send_email_allowed'])
    return render(request, 'serviceform/login/unsubscribe_participant.html',
                  {'participant': participant,
                   'service_form': participant.form})



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from django.utils.translation import gettext_lazy as _

bulk_email_to_responsibles = _("""Dear {{responsible}},

Participation results for {{form}} are now available for you to view.
You can see all participants for the activities you are responsible of in the following URL:
{{url}}
From now on, you will also receive a notification message, if a new participation submitted to
the areas you are responsible of.You can also adjust your contact details and email notification
preferences from previously given URL.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")


invite = _("""Dear {{participant}},

You are invited to participate in "{{ form }}".
You can fill in your participation details at {{ url }}.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")

message_to_responsibles = _("""Dear {{responsible}},

New participation from {{participant}} has just been submitted to {{form}}.
You can see all participants for the activities you are responsible of in the following URL:
{{url}}
You can also adjust your contact details and email notification preferences from that URL.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")

participant_new_form_revision = _("""Dear {{participant}},

New form revision to "{{ form }}" has been published.
Please update your participation information at {{ url }}.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")

participant_on_finish = _("""Dear {{participant}},

You submitted form "{{ form }}" on {{ last_modified }}.
If you wish to change any of the details you gave, you can go to {{ url }}.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")

resend_email_to_participants = _("""Dear {{participant}},

You submitted form "{{ form }}" on {{ last_modified }}.
If you wish to change any of the details you gave, you can go to {{ url }}.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")

participant_on_update = _("""Dear {{participant}},

You submitted update to your data on form "{{ form }}" on {{ last_modified }}.
If you wish to change any of the details you gave, you can go to {{ url }}.

Best regards,
Service form system administrators

Contact person:
{{contact}}""")


request_responsible_auth_link = _("""Dear {{responsible}},

You can see all participants for the activities you are responsible of in the following URL:
{{url}}

Best regards,
Service form system administrators

Contact person:
{{contact}}""")


verification_email_to_participant = _("""Dear {{participant}},

Your email address needs to be verified. Please do so by clicking link below. Then you can
continue filling the form.

{{url}}

Best regards,
Service form system administrators

Contact person:
{{contact}}""")



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, HttpRequest, HttpResponse
from django.shortcuts import render

from .. import forms, models
from ..utils import clean_session
from ..views.decorators import serviceform


@serviceform
def password_login(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    clean_session(request)
    password_form = forms.PasswordForm(service_form)

    if request.method == 'POST':
        password_form = forms.PasswordForm(service_form, request.POST)
        if password_form.is_valid():
            participant = models.Participant.objects.create(
                form_revision=service_form.current_revision)
            request.session['authenticated_participant'] = participant.pk
            return HttpResponseRedirect(reverse('contact_details'))

    return render(request, 'serviceform/login/password_login.html',
                  {'password_form': password_form, 'service_form': service_form})


@serviceform
def send_participant_email(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    email_form = forms.ParticipantSendEmailForm(service_form, request)

    if request.method == 'POST':
        email_form = forms.ParticipantSendEmailForm(service_form, request, request.POST)
        if email_form.is_valid():
            email_form.save()
            return HttpResponseRedirect(
                reverse('send_participant_email', args=(service_form.slug,)))

    return render(request, 'serviceform/login/send_participant_auth_link.html',
                  {'email_form': email_form, 'service_form': service_form})


@serviceform
def send_responsible_email(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    email_form = forms.ResponsibleSendEmailForm(service_form, request)

    if request.method == 'POST':
        email_form = forms.ResponsibleSendEmailForm(service_form, request, request.POST)
        if email_form.is_valid():
            email_form.save()
            return HttpResponseRedirect(
                reverse('send_responsible_email', args=(service_form.slug,)))

    return render(request, 'serviceform/login/send_responsible_auth_link.html',
                  {'email_form': email_form, 'service_form': service_form})



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import logout
from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, Http404, HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, render, redirect
from django.utils.translation import gettext_lazy as _

from .. import models, forms
from ..utils import user_has_serviceform_permission, fetch_participants, expire_auth_link, decode, \
    RevisionOptions
from .decorators import serviceform, require_authenticated_responsible


def authenticate_responsible_old(request: HttpRequest, uuid: str) -> HttpResponse:
    """
    Just expire old and insecure authrorization link if such is being used and send a new one.
    """
    if not uuid:
        raise Http404
    responsible = get_object_or_404(models.ResponsibilityPerson.objects.all(), secret_key=uuid)
    return expire_auth_link(request, responsible)


def authenticate_responsible(request: HttpRequest, responsible_id: int,
                             password: str) -> HttpResponse:
    responsible = get_object_or_404(models.ResponsibilityPerson.objects.all(), pk=responsible_id)
    result = responsible.check_auth_key(password)
    if result == responsible.PasswordStatus.PASSWORD_NOK:
        messages.error(request, _(
            "Given URL might be expired. Please give your email address and we'll send "
            "you a new link"))
        return redirect('send_responsible_email', responsible.form.slug)
    elif result == responsible.PasswordStatus.PASSWORD_EXPIRED:
        return expire_auth_link(request, responsible)

    request.session['authenticated_responsibility'] = responsible.pk
    return HttpResponseRedirect(reverse('responsible_report'))


@login_required(login_url=settings.LOGIN_URL)
def authenticate_responsible_mock(request: HttpRequest, responsible_id: int) -> HttpResponse:
    """
    Mocked authentication to responsible view from admin panel
    """
    responsible = get_object_or_404(models.ResponsibilityPerson.objects.all(), pk=responsible_id)
    user_has_serviceform_permission(request.user, responsible.form, raise_permissiondenied=True)

    request.session['authenticated_responsibility'] = responsible.pk
    return HttpResponseRedirect(reverse('responsible_report'))


@serviceform(check_form_permission=True)
def settings_view(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    form = forms.ReportSettingsForm(service_form, request)
    if request.method == 'POST':
        form = forms.ReportSettingsForm(service_form, request, request.POST)
        if form.is_valid():
            form.save()
            messages.info(request, _('Settings saved'))
        else:
            messages.error(request, _('Settings could not be saved'))
            # return HttpResponseRedirect('') # TODO: why does this not work?
    return render(request, 'serviceform/reports/settings.html',
                  {'service_form': service_form, 'form': form})


@serviceform(check_form_permission=True, init_counters=True)
def all_responsibles(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    return render(request, 'serviceform/reports/all_responsibles.html',
                  {'service_form': service_form})


@serviceform(check_form_permission=True, fetch_participants=True)
def all_participants(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    return render(request, 'serviceform/reports/all_participants.html',
                  {'service_form': service_form})


@serviceform(check_form_permission=True, init_counters=True, fetch_participants=True)
def all_activities(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    return render(request, 'serviceform/reports/all_activities.html',
                  {'service_form': service_form})


@serviceform(check_form_permission=True, init_counters=True, fetch_participants=True)
def all_questions(request: HttpRequest, service_form: models.ServiceForm) -> HttpResponse:
    return render(request, 'serviceform/reports/all_questions.html',
                  {'service_form': service_form})


@require_authenticated_responsible
def view_participant(request: HttpRequest, responsible: models.ResponsibilityPerson,
                     participant_id: int) -> HttpResponse:
    participant = get_object_or_404(models.Participant.objects, pk=participant_id)
    anonymous = False

    if request.user.pk:
        user_has_serviceform_permission(request.user, participant.form)
        user = request.user
    else:
        if responsible.form != participant.form:
            raise PermissionDenied
        user = responsible
        anonymous = True

    if request.method == 'POST':
        form = forms.LogForm(participant, user, request.POST)
        if form.is_valid():
            form.save()
        return HttpResponseRedirect(reverse('view_user', args=(participant_id,)))

    form = forms.LogForm(participant, user)
    service_form = participant.form
    return render(request, 'serviceform/reports/view_participant.html',
                  {'service_form': service_form, 'participant': participant, 'log_form': form,
                   'anonymous': anonymous})


@require_authenticated_responsible
def view_responsible(request: HttpRequest, auth_responsible: models.ResponsibilityPerson,
                     responsible_pk: int) -> HttpResponse:
    responsible = models.ResponsibilityPerson.objects.get(pk=responsible_pk)
    if not (user_has_serviceform_permission(request.user, responsible.form,
                                            raise_permissiondenied=False)
            or (auth_responsible and auth_responsible.show_full_report
                and responsible.form == auth_responsible.form)):
        raise PermissionDenied
    service_form = responsible.form
    request.service_form = service_form
    service_form.init_counters()
    fetch_participants(service_form, revision_name=RevisionOptions.ALL)
    return render(request, 'serviceform/reports/responsible.html',
                  {'service_form': responsible.form, 'responsible': responsible,
                   'show_report_btn': True})


@require_authenticated_responsible
def preview_form(request: HttpRequest, responsible: models.ResponsibilityPerson,
                 slug: str) -> HttpResponse:
    service_form = get_object_or_404(models.ServiceForm.objects, slug=slug)
    user_has_serviceform_permission(request.user, service_form)
    service_form.init_counters()
    form = forms.ParticipationForm(request, None, None, service_form=service_form)
    return render(request, 'serviceform/participation/participation_view.html',
                  {'form': form, 'service_form': service_form, 'readonly': True})


@require_authenticated_responsible
def preview_printable(request: HttpRequest, responsible: models.ResponsibilityPerson,
                      slug: str) -> HttpResponse:
    service_form = get_object_or_404(models.ServiceForm.objects, slug=slug)
    user_has_serviceform_permission(request.user, service_form)
    service_form.init_counters()
    return render(request, 'serviceform/preview_printable.html',
                  {'form': service_form, 'preview': True, 'printable': True})


@require_authenticated_responsible
def edit_responsible(request: HttpRequest,
                     responsible: models.ResponsibilityPerson) -> HttpResponse:
    if responsible is None:
        raise PermissionDenied
    service_form = responsible.form
    form = forms.ResponsibleForm(instance=responsible)
    if request.method == 'POST':
        form = forms.ResponsibleForm(request.POST, instance=responsible)
        if form.is_valid():
            form.save()
            messages.info(request, _('Saved contact details'))
    return render(request, 'serviceform/reports/edit_responsible.html',
                  {'form': form, 'service_form': service_form, 'responsible': responsible})


@require_authenticated_responsible
def responsible_report(request: HttpRequest,
                       responsible: models.ResponsibilityPerson) -> HttpResponse:
    if responsible is None:
        raise PermissionDenied
    service_form = responsible.form
    service_form.init_counters(all_responsibles=True)
    fetch_participants(service_form, revision_name=RevisionOptions.ALL)
    return render(request, 'serviceform/reports/responsible_anonymous.html',
                  {'service_form': responsible.form, 'responsible': responsible})


def logout_view(request: HttpRequest, **kwargs) -> HttpResponse:
    responsible_pk = request.session.pop('authenticated_responsibility', None)
    logout(request)
    messages.info(request, _('You have been logged out'))
    if responsible_pk:
        responsible = models.ResponsibilityPerson.objects.get(pk=responsible_pk)
        return HttpResponseRedirect(reverse('password_login', args=(responsible.form.slug,)))
    return HttpResponseRedirect(reverse('main_page'))


@login_required(login_url=settings.LOGIN_URL)
def invite(request: HttpRequest, serviceform_slug: str, **kwargs) -> HttpResponse:
    service_form = get_object_or_404(models.ServiceForm.objects, slug=serviceform_slug)
    user_has_serviceform_permission(request.user, service_form)

    if request.method == 'POST':
        form = forms.InviteForm(request.POST, instance=service_form)
        if form.is_valid():
            form.save(request=request)
            return HttpResponseRedirect(reverse('invite', args=(service_form.slug,)))
        else:
            return render(request, 'serviceform/reports/invite.html',
                          {'form': form, 'service_form': service_form})
    else:
        form = forms.InviteForm(instance=service_form)
        return render(request, 'serviceform/reports/invite.html',
                      {'form': form, 'service_form': service_form})


@require_authenticated_responsible
def to_full_report(request: HttpRequest, responsible: models.ResponsibilityPerson) -> HttpResponse:
    if not responsible.show_full_report:
        raise PermissionDenied
    return redirect('report', responsible.form.slug)


def unsubscribe(request: HttpRequest, secret_id: str) -> HttpResponse:
    responsible = get_object_or_404(models.ResponsibilityPerson.objects, pk=decode(secret_id))
    responsible.send_email_notifications = False
    responsible.save(update_fields=['send_email_notifications'])
    return render(request, 'serviceform/login/unsubscribe_responsible.html',
                  {'responsible': responsible,
                   'service_form': responsible.form})



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from typing import Iterable
import collections
from django.contrib import admin
from django.conf import settings
from django.contrib import messages
from django.db.models import Model, Field
from django.forms.utils import pretty_name
from django.http import HttpRequest
from django.utils.encoding import force_str
from django.utils.translation import ugettext_lazy as _, gettext as _g
from django.utils import timezone
from guardian.admin import GuardedModelAdmin, GuardedModelAdminMixin
from guardian.shortcuts import get_objects_for_user, assign_perm

from nested_admin.nested import NestedModelAdmin, NestedTabularInline, NestedStackedInline, \
    NestedModelAdminMixin
from . import models, utils

if 'grappelli' in settings.INSTALLED_APPS:
    from grappelli.forms import GrappelliSortableHiddenMixin
else:
    class GrappelliSortableHiddenMixin:
        pass


class OwnerSaveMixin:
    def save_model(self, request: HttpRequest, obj: Model, form, change: bool) -> Model:
        assign_new_perm = not obj.pk
        rv = super().save_model(request, obj, form, change)
        if assign_new_perm:
            assign_perm('serviceform.can_access_serviceform', request.user, obj)
        return rv


class ResponsibleMixin:
    def formfield_for_manytomany(self, db_field: Field, request: HttpRequest=None, **kwargs):
        formfield = super().formfield_for_manytomany(db_field, request, **kwargs)
        if db_field.name in ['responsibles']:
            formfield.queryset = request._responsibles
            # Let's do a little DB optimization trick because there will be
            # otherwise tons of queries to responsible choices
            choices = getattr(request, '_responsible_choices', None)
            formfield.choices = choices or formfield.choices
            if not choices:
                request._responsible_choices = formfield.choices
        return formfield


class ExtendedLogMixin:
    """
    Adds changed content into log message in Admin
    """

    def map_field_value(self, values, choices):
        def do_map(value, choices):
            if value is None:
                return '?'
            elif isinstance(value, Model):
                return str(value)
            else:
                return _g(str(choices.get(value, value)))

        if isinstance(values, collections.Iterable) and not isinstance(values, str):
            msg = ', '.join(do_map(val, choices) for val in values)
            return '[%s]' % msg
        else:
            return do_map(values, choices)

    def get_field_change(self, form, changed_fields):
        """Mentions old and new value for changed fields"""
        changes = []
        for field_name in changed_fields:
            field = form.fields[field_name]
            label = field.label or pretty_name(field_name)
            old_value = field.prepare_value(form.initial.get(field_name))
            new_value = form.cleaned_data[field_name]
            if hasattr(field, 'choices'):
                choices = dict(field.choices)
                old_value = self.map_field_value(old_value, choices)
                new_value = self.map_field_value(new_value, choices)
            if old_value != new_value:
                if isinstance(old_value, datetime.datetime):
                    old_value = old_value.astimezone(timezone.get_current_timezone())
                changes.append({'label': force_str(label).lower(),
                                'new_value': new_value,
                                'old_value': old_value})
        # Translators: This appears after "Changed" and is joined with "and"
        chg_msg = _(' {label} to {new_value} (was {old_value}) ')
        return _('and').join(chg_msg.format(**cc) for cc in changes)

    def construct_change_message(self, request, form, formsets, add=False):
        """Construct a change message from a changed object.
        Copy of contrib/admin/options.py class ModelAdmin with get_text_list
        replaced with self.get_field_change."""
        change_message = []
        if form.changed_data:
            change_message.append(_('Changed %s.')
                                  % self.get_field_change(form, form.changed_data))

        if formsets:
            for formset in formsets:
                form_by_instance = {form.instance: form for form in formset.forms if
                                    form.instance.pk}
                for added_object in formset.new_objects:
                    change_message.append(
                        _('Added %(name)s "%(object)s".')
                        % {'name': force_str(added_object._meta.verbose_name),
                           'object': force_str(added_object)})
                for changed_object, changed_fields in formset.changed_objects:
                    form = form_by_instance[changed_object]
                    change_message.append(
                        _('Changed %(list)s for %(name)s "%(object)s".')
                        % {'list': self.get_field_change(form, changed_fields),
                           'name': force_str(changed_object._meta.verbose_name),
                           'object': force_str(changed_object)})
                for deleted_object in formset.deleted_objects:
                    change_message.append(
                        _('Deleted %(name)s "%(object)s".')
                        % {'name': force_str(deleted_object._meta.verbose_name),
                           'object': force_str(deleted_object)})
        change_message = ' '.join(change_message)
        return change_message or _('No fields changed.')


class ServiceFormItemInline(GrappelliSortableHiddenMixin, ResponsibleMixin, NestedTabularInline):
    extra = 0
    sortable_field_name = 'order'


class ActivityChoiceInline(ServiceFormItemInline):
    fields = ('name', 'description', 'responsibles', 'skip_numbering', 'people_needed', 'order')
    model = models.ActivityChoice


class ActivityInline(ServiceFormItemInline):
    fields = ('name', 'description', 'responsibles', 'multiple_choices_allowed', 'skip_numbering',
              'people_needed', 'order')
    model = models.Activity
    inlines = [ActivityChoiceInline]


class Level2CategoryInline(ServiceFormItemInline):
    fields = ('name', 'description', 'responsibles', 'background_color', 'order')
    model = models.Level2Category
    inlines = [ActivityInline]


class Level1CategoryInline(ServiceFormItemInline):
    fields = ('name', 'description', 'responsibles', 'background_color', 'order')
    model = models.Level1Category
    inlines = [Level2CategoryInline]


class QuestionInline(ResponsibleMixin, GrappelliSortableHiddenMixin, NestedTabularInline):
    sortable_field_name = 'order'
    model = models.Question
    extra = 0
    fields = ('question', 'responsibles', 'answer_type', 'required', 'order')


class EmailTemplateInline(NestedStackedInline):
    fields = ('name', 'subject', 'content')
    model = models.EmailTemplate
    extra = 0


class RevisionInline(NestedStackedInline):
    fields = ('name', ('valid_from', 'valid_to'), 'send_emails_after',
              'send_bulk_email_to_participants')
    model = models.FormRevision
    extra = 0


class ResponsibilityPersonInline(NestedStackedInline):
    model = models.ResponsibilityPerson
    extra = 0
    fields = (('forenames', 'surname'), ('email', 'phone_number'), 'street_address',
              ('postal_code', 'city'), 'send_email_notifications', 'hide_contact_details',
              'show_full_report', 'personal_link')
    readonly_fields = ('personal_link',)


@admin.register(models.ServiceForm)
class ServiceFormAdmin(OwnerSaveMixin, ExtendedLogMixin, NestedModelAdminMixin,
                       GuardedModelAdminMixin, admin.ModelAdmin):
    class Media:
        css = {'all': ('serviceform/serviceform_admin.css',)}

    inlines = [RevisionInline, EmailTemplateInline, ResponsibilityPersonInline,
               Level1CategoryInline, QuestionInline]

    superuser_actions = ['bulk_email_former_participants', 'bulk_email_responsibles']
    if settings.DEBUG:
        superuser_actions.append('shuffle_data')

    actions = superuser_actions

    prepopulated_fields = {'slug': ('name',)}
    list_display = (
        'name', 'id', 'can_access', 'last_editor', 'responsible', 'created_at', 'last_updated',
        'current_revision', 'is_published_display', 'participation_count', 'links')
    list_select_related = ('responsible', 'current_revision', 'last_editor')

    basic = ('name',
             'slug',
             )
    ownership = ('responsible', 'can_access')

    email_settings = (
        'require_email_verification',
        'verification_email_to_participant',
        'email_to_responsibles',
        'email_to_invited_users',

        'email_to_participant',
        'resend_email_to_participant',

        'email_to_participant_on_update',
        'email_to_former_participants',

        'bulk_email_to_responsibles',
        'email_to_responsible_auth_link',
    )

    customization = (
        'current_revision',
        'password',
        'hide_contact_details',
        ('flow_by_categories', 'allow_skipping_categories'),
        'level1_color',
        'level2_color',
        'activity_color',
        'login_text',
        'description',
        'instructions',
    )
    visible_contact_details = (
    ('visible_year_of_birth', 'visible_street_address', 'visible_phone_number'),)
    required_contact_details = (
    ('required_year_of_birth', 'required_street_address', 'required_phone_number'),)

    fieldsets = ((_('Basic information'), {'fields': basic}),
                 (_('Ownership'), {'fields': ownership}),
                 (_('Email settings'), {'fields': email_settings}),
                 (_('Customization'), {'fields': customization}),
                 (_('Ask details from participants'), {'fields': visible_contact_details}),
                 (_('Require details from participants'), {'fields': required_contact_details}),
                 )

    new_fieldsets = ((_('Basic information'), {'fields': basic}),)
    readonly_fields = ('can_access',)

    def get_fieldsets(self, request: HttpRequest, obj: models.ServiceForm=None):
        return self.new_fieldsets if obj is None else super().get_fieldsets(request, obj)

    def get_inline_instances(self, request: HttpRequest, obj: models.ServiceForm=None):
        return super().get_inline_instances(request, obj) if obj else []

    def get_actions(self, request: HttpRequest):
        actions = super().get_actions(request)
        if not request.user.is_superuser:
            for a in self.superuser_actions:
                actions.pop(a, None)
        return actions

    def bulk_email_former_participants(self, request: HttpRequest,
                                       queryset: Iterable[models.ServiceForm]) -> None:
        for serviceform in queryset:
            serviceform.bulk_email_former_participants()

    bulk_email_former_participants.short_description = _('Bulk email former participants now!')

    def bulk_email_responsibles(self, request: HttpRequest,
                                queryset: Iterable[models.ServiceForm]) -> None:
        for serviceform in queryset:
            serviceform.bulk_email_responsibles()

    bulk_email_responsibles.short_description = _('Bulk email responsibility persons now!')

    def shuffle_data(self, request: HttpRequest,
                     queryset: Iterable[models.ServiceForm]) -> None:
        for serviceform in queryset:
            utils.shuffle_person_data(serviceform)

    shuffle_data.short_description = _('Shuffle participant data')

    def get_queryset(self, request: HttpRequest):
        qs = super().get_queryset(request)
        return get_objects_for_user(request.user, 'serviceform.can_access_serviceform', klass=qs,
                                    use_groups=True)

    def get_form(self, request: HttpRequest, obj: models.ServiceForm=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if obj:
            request._responsibles = responsibles = models.ResponsibilityPerson.objects.filter(
                form=obj)
            form.base_fields['responsible'].queryset = responsibles
            form.base_fields['current_revision'].queryset = models.FormRevision.objects.filter(
                form=obj)

            emailtemplates = models.EmailTemplate.objects.filter(form=obj)

            for name, field in form.base_fields.items():
                if 'email_to' in name:
                    field.queryset = emailtemplates
                    if obj and field.queryset:
                        field.required = True

        return form

    def save_model(self, request: HttpRequest, obj: models.ServiceForm, form, change: bool):
        obj.last_editor = request.user
        rv = super().save_model(request, obj, form, change)
        if not change:
            obj.create_initial_data()
        if obj.pk:
            count = obj.level1category_set.count()
            if obj.flow_by_categories and count > 7:
                messages.warning(request, _(
                    'You have {} level 1 categories. We recommend that no more than '
                    '7 level 1 categories are used, if form flow is split to categories, '
                    'so that form is rendered correctly.').format(count))
        return rv

    def save_related(self, request: HttpRequest, form, formsets, change: bool):
        rv = super().save_related(request, form, formsets, change)
        form.instance.reschedule_bulk_email()
        return rv


@admin.register(models.EmailMessage)
class EmailMessageAdmin(ExtendedLogMixin, admin.ModelAdmin):
    list_display = ('to_address', 'created_at', 'sent_at', 'subject_display', 'template',
                    'content_display',)


@admin.register(models.Participant)
class ParticipantAdmin(ExtendedLogMixin, admin.ModelAdmin):
    list_display = (
        'id', '__str__', 'form_display', 'form_revision', 'status', 'activities_display',
        'created_at', 'last_modified', 'personal_link')
    fields = ('forenames', 'surname')

    def get_queryset(self, request):
        qs = super().get_queryset(request).select_related('form_revision__form')
        if request.user.is_superuser or settings.OTHER_CAN_SEE_DATA:
            return qs
        allowed_forms = get_objects_for_user(request.user, 'serviceform.can_access_serviceform',
                                             models.ServiceForm)
        return qs.filter(form_revision__form__in=allowed_forms).distinct()



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import string
import logging
from enum import Enum
from typing import Tuple, Set, Optional, Sequence, Iterator, Iterable, TYPE_CHECKING

from colorful.fields import RGBColorField
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.db.models import Prefetch
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from guardian.shortcuts import get_users_with_perms
from select2 import fields as select2_fields

from serviceform.tasks.models import Task

from .. import emails, utils
from ..utils import ColorStr

from .mixins import SubitemMixin, NameDescriptionMixin, CopyMixin
from .people import Participant, ResponsibilityPerson
from .email import EmailTemplate
from .participation import QuestionAnswer

if TYPE_CHECKING:
    from .participation import ParticipationActivity, ParticipationActivityChoice

local_tz = timezone.get_default_timezone()
logger = logging.getLogger(__name__)


class ColorField(RGBColorField):
    def get_prep_value(self, value: 'ColorStr') -> 'Optional[ColorStr]':
        rv = super().get_prep_value(value)
        if rv == '#000000':
            rv = None
        return rv

    def from_db_value(self, value: 'Optional[ColorStr]', *args):
        if value is None:
            return '#000000'
        return value



class FormRevision(models.Model):
    class Meta:
        verbose_name = _('Form revision')
        verbose_name_plural = _('Form revisions')
        ordering = ('-valid_from',)
        unique_together = ('form', 'name')

    name = models.SlugField(max_length=32, verbose_name=_('Revision name'), db_index=True)
    form = models.ForeignKey('ServiceForm', verbose_name=_('Service form'))
    valid_from = models.DateTimeField(verbose_name=_('Valid from'),
                                      default=datetime.datetime(3000, 1, 1, tzinfo=local_tz))
    valid_to = models.DateTimeField(verbose_name=_('Valid to'),
                                    default=datetime.datetime(3000, 1, 1, tzinfo=local_tz))
    send_bulk_email_to_participants = models.BooleanField(
        _('Send bulk email to participants'),
        help_text=_('Send email to participants that filled the form when this revision was '
                    'active. Email is sent when new current revision is published.'),
        default=True)
    send_emails_after = models.DateTimeField(
        verbose_name=_('Email sending starts'),
        help_text=_(
            'Sends bulk email to responsibility persons at specified time, after which it will '
            'send email for each new participation'),
        default=datetime.datetime(3000, 1, 1, tzinfo=local_tz))

    def __str__(self):
        return self.name


class ServiceForm(SubitemMixin, models.Model):
    subitem_name = 'level1category'

    class Meta:
        verbose_name = _('Service form')
        verbose_name_plural = _('Service forms')

    def __str__(self):
        return self.name

    # Basic info
    name = models.CharField(max_length=256, verbose_name=_('Name of the serviceform'))
    slug = models.SlugField(unique=True, verbose_name=_('Slug'), db_index=True, help_text=_(
        'This is part of the form url, i.e. form will be located {}/yourslug').format(
        settings.SERVER_URL))

    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Created at'))
    last_updated = models.DateTimeField(auto_now=True, null=True, verbose_name=_('Last updated'))
    last_editor = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_('Last edited by'),
                                    related_name='last_edited_serviceform', null=True,
                                    on_delete=models.SET_NULL)

    # Ownership
    responsible = models.ForeignKey(ResponsibilityPerson, null=True, blank=True,
                                    verbose_name=_('Responsible'), on_delete=models.SET_NULL)

    # Email settings
    require_email_verification = models.BooleanField(_('Require email verification'), default=True)

    verification_email_to_participant = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Verification email to participant'),
        help_text=_(
            'Email verification message that is sent to participant when filling form, '
            'if email verification is enabled'),
        on_delete=models.SET_NULL)

    email_to_responsibles = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Email to responsibles'), help_text=_(
            'Email that is sent to responsibles when new participation is registered'),
        on_delete=models.SET_NULL)

    bulk_email_to_responsibles = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        verbose_name=_('Bulk email to responsibles'),
        help_text=_('Email that is sent to responsibles when emailing starts'),
        related_name='+',
        on_delete=models.SET_NULL)

    email_to_responsible_auth_link = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Responsible requests '
                       'auth link'),
        help_text=_('Email that is sent to responsible when he requests auth link'),
        on_delete=models.SET_NULL)

    # Participant emails:

    # on_finish
    email_to_participant = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Email to participant, on finish'),
        help_text=_('Email that is sent to participant after he has fulfilled his participation'),
        on_delete=models.SET_NULL)
    # on update
    email_to_participant_on_update = models.ForeignKey(EmailTemplate, null=True, blank=True,
                                                       related_name='+', verbose_name=_(
            'Email to participant, on update'), help_text=_(
            'Email that is sent to participant after he has updated his participation'),
                                                       on_delete=models.SET_NULL)
    # resend
    resend_email_to_participant = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Resend email to participant'),
        help_text=_('Email that is sent to participant if he requests resending email'),
        on_delete=models.SET_NULL)
    # new_form_revision
    email_to_former_participants = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Bulk email to former participants'),
        help_text=_('Email that is sent to former participants when form is published'),
        on_delete=models.SET_NULL)
    # invite
    email_to_invited_users = models.ForeignKey(
        EmailTemplate, null=True, blank=True,
        related_name='+',
        verbose_name=_('Invite email'),
        help_text=_(
            'Email that is sent when user is invited to the form manually via invite form'),
        on_delete=models.SET_NULL)

    # Form settings
    current_revision = models.ForeignKey(
        FormRevision, null=True, blank=True,
        verbose_name=_('Current revision'),
        help_text=_(
            'You need to first add a revision to form (see below) and save. '
            'Then newly created revision will appear in the list.'),
        on_delete=models.SET_NULL)

    password = models.CharField(
        _('Password'), max_length=32, blank=True,
        help_text=_('Password that is asked from participants'),
        default='')

    hide_contact_details = models.BooleanField(
        _('Hide contact details (other than email) in form'), default=False)
    flow_by_categories = models.BooleanField(_('Split participation form to level 1 categories'),
                                             default=False, help_text=_(
            'Please note that preview shows full form despite this option'))
    allow_skipping_categories = models.BooleanField(
        _('Allow jumping between categories'),
        default=False,
        help_text=_('In effect only if flow by categories option is enabled. If this option is '
                    'enabled, user can jump between categories. If disabled, he must proceed them '
                    'one by one.'))

    level1_color = ColorField(_('Level 1 category default background color'), null=True,
                              blank=True,
                              help_text=_('If left blank (black), default coloring will be used'))
    level2_color = ColorField(_('Level 2 category default background color'), null=True,
                              blank=True, help_text=_(
            'If left blank (black), it will be derived from level 1 background color'))
    activity_color = ColorField(_('Activity default background color'), null=True, blank=True,
                                help_text=_('If left blank (black), it will be derived from '
                                'level 2 background color'))

    description = models.TextField(blank=True, verbose_name=_('Description'), help_text=_(
        'Description box will be shown before instruction box in participation view.'))
    instructions = models.TextField(
        _('Instructions'), help_text=_(
            'Use HTML formatting. Leave this empty to use default. '
            'This is shown in participation view.'),
        blank=True, null=True)
    login_text = models.TextField(_('Login text'), blank=True, null=True,
                                  help_text=_('This will be shown in the login screen'))

    required_year_of_birth = models.BooleanField(_('Year of birth'), default=False)
    required_street_address = models.BooleanField(_('Street address'), default=True)
    required_phone_number = models.BooleanField(_('Phone number'), default=True)

    visible_year_of_birth = models.BooleanField(_('Year of birth'), default=True)
    visible_street_address = models.BooleanField(_('Street address'), default=True)
    visible_phone_number = models.BooleanField(_('Phone number'), default=True)

    tasks = GenericRelation(Task, object_id_field='target_id', content_type_field='target_type')

    class InviteUserResponse(Enum):
        EMAIL_SENT = 0
        USER_DENIED_EMAIL = 1
        USER_EXISTS = 2

    def can_access(self) -> str:
        return ', '.join('%s' % u for u in get_users_with_perms(self))

    can_access.short_description = _('Can access')

    @cached_property
    def sub_items(self) -> 'Sequence[AbstractServiceFormItem]':
        lvl2s = Prefetch('level2category_set',
                         queryset=Level2Category.objects.prefetch_related('responsibles'))
        acts = Prefetch('level2category_set__activity_set',
                        queryset=Activity.objects.prefetch_related('responsibles'))
        choices = Prefetch('level2category_set__activity_set__activitychoice_set',
                           queryset=ActivityChoice.objects.prefetch_related('responsibles'))

        return self.level1category_set.prefetch_related('responsibles').prefetch_related(
            lvl2s, acts, choices)

    def create_initial_data(self) -> None:
        self.create_email_templates()
        self.current_revision = FormRevision.objects.create(name='%s' % timezone.now().year,
                                                            form=self)
        self.responsible = ResponsibilityPerson.objects.create(
            forenames=_('Default'),
            surname=_('Responsible'),
            email=_('defaultresponsible@email.com'),
            form=self)
        self.save()

    def create_email_templates(self) -> None:
        if not self.pk:
            logger.error('Cannot create email template if form is not saved')
            return

        commit = False
        # TODO: refactor this
        if not self.bulk_email_to_responsibles:
            commit = True
            self.bulk_email_to_responsibles = EmailTemplate.make(
                _('Default bulk email to responsibles'), self,
                emails.bulk_email_to_responsibles,
                _('Participations can be now viewed for form {{form}}'))
        if not self.email_to_responsibles:
            commit = True
            self.email_to_responsibles = EmailTemplate.make(
                _('Default email to responsibles'),
                self, emails.message_to_responsibles,
                _('New participation arrived for form {{form}}'))
        if not self.email_to_participant:
            commit = True
            self.email_to_participant = EmailTemplate.make(
                _('Default email to participant, on finish'), self,
                emails.participant_on_finish,
                _('Your update to form {{form}}'))
        if not self.email_to_participant_on_update:
            commit = True
            self.email_to_participant_on_update = EmailTemplate.make(
                _('Default email to participant, on update'), self,
                emails.participant_on_update,
                _('Your updated participation to form {{form}}'))
        if not self.email_to_former_participants:
            commit = True
            self.email_to_former_participants = EmailTemplate.make(
                _('Default email to former participants'), self,
                emails.participant_new_form_revision,
                _('New form revision to form {{form}} has been published'))
        if not self.resend_email_to_participant:
            commit = True
            self.resend_email_to_participant = EmailTemplate.make(
                _('Default resend email to participant'), self,
                emails.resend_email_to_participants,
                _('Your participation to form {{form}}'))
        if not self.email_to_invited_users:
            commit = True
            self.email_to_invited_users = EmailTemplate.make(
                _('Default invite email to participants'), self,
                emails.invite,
                _('Invitation to fill participation in {{form}}'))
        if not self.email_to_responsible_auth_link:
            commit = True
            self.email_to_responsible_auth_link = EmailTemplate.make(
                _('Default request responsible auth link email'), self,
                emails.request_responsible_auth_link,
                _('Your report in {{form}}'))
        if not self.verification_email_to_participant:
            commit = True
            self.verification_email_to_participant = EmailTemplate.make(
                _('Default verification email to participant'), self,
                emails.verification_email_to_participant,
                _('Please verify your email in {{form}}'))
        if commit:
            self.save()

    def invite_user(self, email: str, old_participants: bool=False) -> InviteUserResponse:
        """
            Create new participations to current form version and send invites

        :return: int (one of InviteUserResponse constants)
        """
        logger.info('Invite user %s %s', self, email)

        participant = Participant.objects.filter(email=email, form_revision__form=self).first()
        if participant:
            if old_participants and participant.form_revision != self.current_revision:
                rv = participant.send_participant_email(Participant.EmailIds.INVITE)
                return (self.InviteUserResponse.EMAIL_SENT
                        if rv else self.InviteUserResponse.USER_DENIED_EMAIL)
            else:
                return self.InviteUserResponse.USER_EXISTS
        else:
            participant = Participant.objects.create(email=email,
                                                     form_revision=self.current_revision,
                                                     status=Participant.STATUS_INVITED)
            participant.send_participant_email(Participant.EmailIds.INVITE)
            return self.InviteUserResponse.EMAIL_SENT

    @cached_property
    def questions(self) -> 'Sequence[Question]':
        return self.question_set.all()

    def activities(self) -> 'Iterator[Activity]':
        for c1 in self.sub_items:
            for c2 in c1.sub_items:
                for a in c2.sub_items:
                    yield a

    @property
    def is_published(self) -> bool:
        return (self.current_revision and
                self.current_revision.valid_from <= timezone.now()
                                                 <= self.current_revision.valid_to)

    def is_published_display(self) -> bool:
        return self.is_published

    is_published_display.boolean = True
    is_published_display.short_description = _('Is open?')

    def init_counters(self, all_responsibles: bool=True) -> None:
        """
        Initializes counters and collects responsibles from subitems
        """
        if getattr(self, '_counters_initialized', None):
            logger.error('Counters already initialized')
            return
        utils.init_serviceform_counters(self, all_responsibles)
        self._counters_initialized = True

    def _find_new_slug(self) -> str:
        slug = self.slug
        while ServiceForm.objects.filter(slug=slug).exists():
            slug += '-copy'
        return slug

    def links(self) -> Tuple[str]:
        return (format_html('<a href="{}">{}</a>, ', reverse('report', args=(self.slug,)),
                            _('To report')) +
                format_html('<a href="{}">{}</a>, ', reverse('password_login', args=(self.slug,)),
                            _('To form')) +
                format_html('<a href="{}">{}</a>, ', reverse('preview_form', args=(self.slug,)),
                            _('Preview')) +
                format_html('<a href="{}">{}</a>, ',
                            reverse('preview_printable', args=(self.slug,)), _('Printable')) +
                format_html('<a href="{}">{}</a>', reverse('invite', args=(self.slug,)),
                            _('Invite'))
                )

    links.short_description = _('Links')

    def participation_count(self) -> str:
        if self.current_revision:
            old_time = timezone.now() - datetime.timedelta(minutes=20)
            ready = self.current_revision.participant_set.filter(
                status__in=Participant.READY_STATUSES)
            recent_ongoing = self.current_revision.participant_set.filter(
                status__in=[Participant.STATUS_ONGOING],
                last_modified__gt=old_time)

            return '%s + %s' % (ready.count(), recent_ongoing.count())
        else:
            return '0'

    participation_count.short_description = _('Participation count')

    def bulk_email_responsibles(self) -> None:
        logger.info('Bulk email responsibles %s', self)

        for r in self.responsibilityperson_set.all():
            r.send_bulk_mail()

    def bulk_email_former_participants(self) -> None:
        logger.info('Bulk email former participants %s', self)
        for p in Participant.objects.filter(send_email_allowed=True,
                                            form_revision__send_bulk_email_to_participants=True,
                                            form_revision__form=self,
                                            form_revision__valid_to__lt=timezone.now(),
                                            status__in=Participant.READY_STATUSES
                                            ).distinct():
            p.send_participant_email(Participant.EmailIds.NEW_FORM_REVISION)

    def reschedule_bulk_email(self) -> None:
        now = timezone.now()
        self.tasks.filter(scheduled_time__gt=now, status=Task.REQUESTED).delete()

        if not self.current_revision:
            return

        self.current_revision.refresh_from_db()

        if self.current_revision.send_emails_after > now:
            tr = Task.make(self.bulk_email_responsibles,
                           scheduled_time=self.current_revision.send_emails_after)
        if self.current_revision.valid_from > now:
            tp = Task.make(self.bulk_email_former_participants,
                           scheduled_time=self.current_revision.valid_from)


class AbstractServiceFormItem(models.Model):
    _responsibles: Set[ResponsibilityPerson]
    sub_items: 'Iterable[AbstractServiceFormItem]'

    class Meta:
        abstract = True
        ordering = ('order',)

    order = models.PositiveIntegerField(default=0, blank=False, null=False, db_index=True,
                                        verbose_name=_('Order'))
    responsibles = select2_fields.ManyToManyField(ResponsibilityPerson, blank=True,
                                                  verbose_name=_('Responsible persons'),
                                                  related_name='%(class)s_related',
                                                  overlay=_('Choose responsibles'),
                                                  )

    @cached_property
    def responsibles_display(self) -> str:
        first_resp = ''
        responsibles = self.responsibles.all()
        if responsibles:
            first_resp = str(self.responsibles.first())
        if len(responsibles) > 1:
            return _('{} (and others)').format(first_resp)
        else:
            return first_resp

    def background_color_display(self) -> 'ColorStr':
        raise NotImplementedError


class Level1Category(SubitemMixin, NameDescriptionMixin, AbstractServiceFormItem):
    subitem_name = 'level2category'
    background_color = ColorField(_('Background color'), blank=True, null=True)

    class Meta(AbstractServiceFormItem.Meta):
        verbose_name = _('Level 1 category')
        verbose_name_plural = _('Level 1 categories')

    form = models.ForeignKey(ServiceForm, on_delete=models.CASCADE)

    @cached_property
    def background_color_display(self) -> 'ColorStr':
        return utils.not_black(self.background_color) or utils.not_black(self.form.level1_color)


class Level2Category(SubitemMixin, NameDescriptionMixin, AbstractServiceFormItem):
    subitem_name = 'activity'
    background_color = ColorField(_('Background color'), blank=True, null=True)

    class Meta(AbstractServiceFormItem.Meta):
        verbose_name = _('Level 2 category')
        verbose_name_plural = _('Level 2 categories')

    category = models.ForeignKey(Level1Category, null=True, verbose_name=_('Level 1 category'),
                                 on_delete=models.CASCADE)

    @cached_property
    def background_color_display(self) -> 'ColorStr':
        return (utils.not_black(self.background_color) or
                (self.category.background_color_display and utils.lighter_color(
                    self.category.background_color_display)) or
                utils.not_black(self.category.form.level2_color))


class Activity(SubitemMixin, NameDescriptionMixin, AbstractServiceFormItem):
    subitem_name = 'activitychoice'

    class Meta(AbstractServiceFormItem.Meta):
        verbose_name = _('Activity')
        verbose_name_plural = _('Activities')

    category = models.ForeignKey(Level2Category, verbose_name=_('Category'),
                                 on_delete=models.CASCADE)
    multiple_choices_allowed = models.BooleanField(default=True, verbose_name=_('Multichoice'))
    people_needed = models.PositiveIntegerField(_('Needed'), default=0)
    skip_numbering = models.BooleanField(_('Skip'), default=False)

    @property
    def has_choices(self) -> bool:
        return self.activitychoice_set.exists()

    @property
    def id_display(self) -> str:
        return '%s+' % max(1, self._counter) if self.skip_numbering else self._counter

    def participation_items(self, revision_name: str) -> 'Sequence[ParticipationActivity]':
        current_revision = self.category.category.form.current_revision

        qs = self.participationactivity_set.filter(
            participant__status__in=Participant.READY_STATUSES)

        if revision_name == utils.RevisionOptions.ALL:
            qs = qs.order_by('participant__form_revision')
        elif revision_name == utils.RevisionOptions.CURRENT:
            qs = qs.filter(participant__form_revision=current_revision)
        else:
            qs = qs.filter(participant__form_revision__name=revision_name)
        return qs

    @property
    def show_checkbox(self) -> bool:
        has_choices = self.has_choices
        return not has_choices or (has_choices and not self.multiple_choices_allowed)

    @cached_property
    def background_color_display(self) -> 'ColorStr':
        return self.category.background_color_display and utils.lighter_color(
            self.category.background_color_display)


class ActivityChoice(SubitemMixin, NameDescriptionMixin, AbstractServiceFormItem):
    class Meta(AbstractServiceFormItem.Meta):
        verbose_name = _('Activity choice')
        verbose_name_plural = _('Activity choices')

    activity = models.ForeignKey(Activity, on_delete=models.CASCADE)
    people_needed = models.PositiveIntegerField(_('Needed'), default=0)
    skip_numbering = models.BooleanField(_('Skip'), default=False)

    @property
    def id_display(self) -> str:
        letter = string.ascii_lowercase[max(0, self._counter - 1)]
        return '%s+' % letter if self.skip_numbering else letter

    @property
    def is_first(self) -> bool:
        return self._counter == 0

    def participation_items(self, revision_name: str) -> 'Sequence[ParticipationActivityChoice]':
        current_revision = self.activity.category.category.form.current_revision

        qs = self.participationactivitychoice_set.filter(
            activity__participant__status__in=Participant.READY_STATUSES)

        if revision_name == utils.RevisionOptions.ALL:
            qs = qs.order_by('activity__participant__form_revision')
        elif revision_name == utils.RevisionOptions.CURRENT:
            qs = qs.filter(activity__participant__form_revision=current_revision)
        else:
            qs = qs.filter(activity__participant__form_revision__name=revision_name)
        return qs

    @cached_property
    def background_color_display(self) -> 'ColorStr':
        return self.activity.category.background_color_display and utils.lighter_color(
            self.activity.category.background_color_display)


class Question(CopyMixin, AbstractServiceFormItem):
    class Meta(AbstractServiceFormItem.Meta):
        verbose_name = _('Question')
        verbose_name_plural = _('Questions')

    ANSWER_INT = 'integer'
    ANSWER_SHORT_TEXT = 'short_text'
    ANSWER_LONG_TEXT = 'long_text'
    ANSWER_BOOL = 'boolean'
    ANSWER_DATE = 'date'

    ANSWER_TYPES = ((ANSWER_INT, _('Integer')),
                    (ANSWER_SHORT_TEXT, _('Short text')),
                    (ANSWER_LONG_TEXT, _('Long text')),
                    (ANSWER_BOOL, _('Boolean')),
                    (ANSWER_DATE, _('Date')),
                    )

    form = models.ForeignKey(ServiceForm, on_delete=models.CASCADE)
    question = models.CharField(max_length=1024, verbose_name=_('Question'))
    answer_type = models.CharField(max_length=16, choices=ANSWER_TYPES, default=ANSWER_SHORT_TEXT,
                                   verbose_name=_('Answer type'))
    required = models.BooleanField(default=False, verbose_name=_('Answer required?'))

    def render(self) -> str:
        return render_to_string(
            'serviceform/participation/question_form/types/question_%s.html' % self.answer_type,
            {'question': self})

    def questionanswers(self, revision_name: str) -> 'Sequence[QuestionAnswer]':
        qs = QuestionAnswer.objects.filter(question=self,
                                           participant__status__in=Participant.READY_STATUSES)

        current_revision = self.form.current_revision

        if revision_name == utils.RevisionOptions.ALL:
            qs = qs.order_by('-participant__form_revision')
        elif revision_name == utils.RevisionOptions.CURRENT:
            qs = qs.filter(participant__form_revision=current_revision)
        else:
            qs = qs.filter(participant__form_revision__name=revision_name)
        return qs

    def __str__(self):
        return self.question





# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.


import re
import signal
import uuid
import random
import string
import logging
from itertools import chain
from typing import Match, Optional, TYPE_CHECKING, Iterable, Union

if TYPE_CHECKING:
    from .models import ServiceForm, Participant, ResponsibilityPerson, AbstractServiceFormItem

from colorful.forms import RGB_REGEX
from django.contrib import messages
from django.core.cache import caches
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect
from django.utils.safestring import mark_safe

from collections import defaultdict

from colour import Color
from django.core.exceptions import PermissionDenied
from django.utils.translation import ugettext_lazy as _
from django.utils import translation
from django.conf import settings

from django.db import transaction

logger = logging.getLogger(__name__)


class DelayedKeyboardInterrupt(object):
    def __init__(self):
        self.will_interrupt = False
        signal.signal(signal.SIGINT, self.handler)

    def __enter__(self):
        return self

    def handler(self, signum, frame):
        self.will_interrupt = True

    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        if self.will_interrupt:
            raise KeyboardInterrupt()


def _get_ident(request: HttpRequest) -> str:
    service_form = getattr(request, 'service_form', '')
    if not service_form:
        logger.error('No serviceform in _get_ident!?')

    if request.user.pk:
        ident = 'user_%s' % request.user.pk
    else:
        resp = get_responsible(request)
        ident = 'responsible_%s' % resp.pk if resp else 'anonymous'
        service_form = resp.form if resp else None

    return f"{ident}_serviceform_{getattr(service_form, 'pk', '')}"


class RevisionOptions:
    CURRENT = '__current'
    ALL = '__all'


settings_defaults = {'revision': RevisionOptions.CURRENT}


def get_report_settings(request: HttpRequest, parameter: str=None) -> Union[dict, str]:
    cache = caches['persistent']
    report_settings = cache.get('settings_for_%s' % _get_ident(request), settings_defaults.copy())
    if parameter:
        return report_settings.get(parameter)
    return report_settings


def set_report_settings(request: HttpRequest, report_settings: dict) -> None:
    cache = caches['persistent']
    cache.set('settings_for_%s' % _get_ident(request), report_settings)


def user_has_serviceform_permission(user: settings.AUTH_USER_MODEL, service_form: 'ServiceForm',
                                    raise_permissiondenied: bool=True):
    if user.has_perm('serviceform.can_access_serviceform', service_form):
        return True
    else:
        if raise_permissiondenied:
            raise PermissionDenied(_('User is not allowed to access document'))
        else:
            return False


_participants = {}


def get_participant(_id: int) -> 'Participant':
    p = _participants.get(_id)
    if p is None:
        logger.error('Participant %d was not in cache!', _id)
    return p


def fetch_participants(service_form: 'ServiceForm', revision_name: str) -> None:
    global _participants
    from .models import Participant
    is_all_revisions = revision_name == RevisionOptions.ALL
    is_current_revision = revision_name == RevisionOptions.CURRENT

    qs = Participant.objects.prefetch_related('participantlog_set__written_by')
    if is_all_revisions:
        qs = qs.select_related('form_revision')
        participants = qs.filter(form_revision__form=service_form).distinct()
    elif is_current_revision:
        participants = qs.filter(form_revision=service_form.current_revision)
    else:
        participants = qs.filter(form_revision__name=revision_name)

    _participants = {itm.pk: itm for itm in participants}
    return


class ClearParticipantCacheMiddleware:
    def process_request(self, request: HttpRequest):
        _participants.clear()
        _responsible_counts.clear()


class InvalidateCachalotAfterEachRequestMiddleware(object):
    """
    This middleware clears the cachalot cache at the end of every request.
    """

    def process_exception(self, request: HttpRequest, exception: Exception):
        if 'cachalot' in settings.INSTALLED_APPS:
            cachalot_cache = settings.CACHALOT_CACHE
            caches[cachalot_cache].clear()

    def process_response(self, request: HttpRequest, response: HttpResponse):
        if 'cachalot' in settings.INSTALLED_APPS:
            cachalot_cache = settings.CACHALOT_CACHE
            caches[cachalot_cache].clear()
        return response

_responsible_counts = defaultdict(int)


def init_serviceform_counters(service_form: 'ServiceForm', all_responsibles: bool=True) -> None:
    """
    Initializes counters and collects responsibles from subitems

    :return:
    """
    activity_count = 0
    cat1_counter = 0
    _responsible_counts.clear()

    def _add_responsible(responsibles: 'Iterable[ResponsibilityPerson]',
                         *targets: 'AbstractServiceFormItem',
                         resp_count: bool=False) -> None:
        if resp_count:
            for r in {resp for target in targets for resp in target.responsibles.all() if resp}:
                _responsible_counts[r.pk] += 1
        for resp in responsibles:
            for t in targets:
                t._responsibles.add(resp)

    for cat1 in service_form.sub_items:
        cat2_counter = 0
        cat1._counter = cat1_counter
        cat1_counter += 1
        cat1._responsibles = set(cat1.responsibles.all())
        for cat2 in cat1.sub_items:
            cat2._counter = cat2_counter
            cat2_counter += 1
            _add_responsible(cat2.responsibles.all(), cat1, cat2)
            if all_responsibles:
                cat2._responsibles.update(set(cat1.responsibles.all()))
            for activity in cat2.sub_items:
                if not activity.skip_numbering:
                    activity_count += 1
                activity._counter = activity_count

                choice_counter = 0
                _add_responsible(activity.responsibles.all(), cat1, cat2, activity,
                                 resp_count=True)
                if all_responsibles:
                    activity._responsibles.update(
                        set(cat1.responsibles.all()) | set(cat2.responsibles.all()))
                for choice in activity.sub_items:
                    if not choice.skip_numbering:
                        choice_counter += 1
                    choice._counter = choice_counter
                    _add_responsible(choice.responsibles.all(), cat1, cat2, activity, choice,
                                     resp_count=True)
                    if all_responsibles:
                        choice._responsibles.update(set(activity.responsibles.all()) |
                                                    set(cat1.responsibles.all()) |
                                                    set(cat2.responsibles.all()))


def shuffle_person_data(service_form: 'ServiceForm') -> None:
    from .models import Participant, ResponsibilityPerson, Question
    letters = len(string.ascii_letters)
    forenames = set()
    surnames = set()
    participants = Participant.objects.filter(form_revision__form=service_form)
    responsibles = ResponsibilityPerson.objects.filter(form=service_form)
    for p in chain(participants, responsibles):
        for n in p.forenames.split(' '):
            if n:
                forenames.add(n.title())
        for n in p.surname.split('-'):
            if n:
                surnames.add(n.title())

    def shuffle(m, *attrs):
        for a in attrs:
            old = getattr(m, a)
            if old:
                new = [string.ascii_letters[random.randrange(0, letters)] for i in range(len(old))]
                for i in range(len(old)):
                    if old[i] in '@ .,':
                        new[i] = old[i]
                setattr(m, a, ''.join(new))
        m.save()

    def shuffle_question(q):
        if q.question.answer_type in [Question.ANSWER_LONG_TEXT,
                                      Question.ANSWER_SHORT_TEXT]:
            shuffle(q, 'answer')

    forenames = tuple(forenames)
    surnames = tuple(surnames)

    def valid_email(s):
        return re.sub('[^a-zA-Z@\.-]', '', s)

    def shuffle_contact_details(p):
        p.forenames = ' '.join(
            forenames[random.randrange(0, len(forenames))] for i in range(random.randint(1, 2)))
        p.surname = '-'.join(
            surnames[random.randrange(0, len(surnames))] for i in range(random.randint(1, 2)))
        if p.email:
            p.email = valid_email(
                '%s.%s@email.com' % (p.forenames.replace(' ', '.').lower(), p.surname.lower()))
        if p.street_address:
            p.street_address = 'Kontaktikatu %d' % random.randint(0, 99)
        if p.postal_code:
            p.postal_code = ''.join('%s' % random.randint(0, 9) for i in range(5))
        if p.phone_number:
            p.phone_number = ''.join('%s' % random.randint(0, 9) for i in range(9))
        if p.city:
            p.city = 'Hemilä'
        p.save()

    for p in chain(participants, responsibles):
        shuffle_contact_details(p)

    for p in participants:
        for a in p.activities:
            shuffle(a, 'additional_info')
            for c in a.choices:
                shuffle(c, 'additional_info')
        for q in p.questions:
            shuffle_question(q)


def count_for_responsible(resp: 'ResponsibilityPerson') -> int:
    return _responsible_counts[resp.pk]


def generate_uuid() -> str:
    return str(uuid.uuid4())

ColorStr = str  # TODO: Type validation against RGB_REGEX.pattern?


def darker_color(color: ColorStr) -> ColorStr:
    c = Color(color)
    h, s, l = c.get_hsl()
    l_new = l * 0.75  # l - (1.0-l)*0.5
    c.set_hsl((h, s, l_new))
    return c.get_hex()


def lighter_color(color: ColorStr) -> ColorStr:
    c = Color(color)
    h, s, l = c.get_hsl()
    l_new = l + (1.0 - l) * 0.5
    c.set_hsl((h, s, l_new))
    return c.get_hex()


def not_black(color: ColorStr) -> Optional[ColorStr]:
    return color if color != '#000000' else None


def color_for_count(count: int) -> ColorStr:
    if not count:
        return Color('white').get_hex()
    c = Color('green')
    max = 10
    count_real = min(max, count)
    c.hue = (max - count_real) / max * 0.33
    c.saturation = 1.0
    c.luminance = 0.7
    return c.get_hex()


def update_serviceform_default_emails() -> None:
    from .models import ServiceForm
    translation.activate('fi')
    for s in ServiceForm.objects.all():
        s.create_email_templates()


def clean_session(request: HttpRequest):
    keys = ['max_category', 'authenticated_participant', 'authenticated_responsibility',
            'verification_sent']
    for key in keys:
        request.session.pop(key, None)
        # request.session.clear()


def get_responsible(request: HttpRequest):
    from .models import ResponsibilityPerson
    responsible_pk = request.session.get('authenticated_responsibility')
    return ResponsibilityPerson.objects.filter(pk=responsible_pk).first()


def safe_join(sep: str, args_generator: Iterable[str]):
    sep = mark_safe(sep)
    result = mark_safe('')
    args = list(args_generator)
    for a in args[:-1]:
        result += a
        result += sep
    result += args[-1]
    return result


def expire_auth_link(request: HttpRequest, obj: 'Union[Participant, ResponsibilityPerson]') \
        -> HttpResponse:
    """

    :param request: WSGI request
    :param obj: either Participant or ResponsibilityPerson
    :return: HttpResponse
    """
    obj.resend_auth_link()
    messages.info(request,
                  _('Your authentication URL was expired. New link has been sent to {}').format(
                      obj.email))
    return redirect('password_login', obj.form.slug)


def encode(number: int) -> str:
    letters = settings.CODE_LETTERS
    result = ''.join(reversed('%o' % (100000 + number)))
    for ii in range(0, 9, 2):
        result = result.replace(str(ii), letters[ii // 2], 1)
    return result


def decode(number: str) -> Optional[int]:
    letters = settings.CODE_LETTERS
    if number is None:
        return None
    for ii in range(0, 9, 2):
        number = number.replace(letters[ii // 2], str(ii))
    try:
        result = int(''.join(reversed(number)), 8)
    except ValueError:
        return None
    result -= 100000
    if result < 0:
        result = None
    return result




from typing import NamedTuple, Dict, List, TYPE_CHECKING, Union, Iterable, Sequence

from django import template
from django.core.urlresolvers import reverse
from django.template import Context
from django.utils.html import format_html
from django.utils.safestring import mark_safe, SafeString
from django.utils.translation import gettext_lazy as _

from ..models import Participant
from ..utils import safe_join, ColorStr
from .. import utils
from ..urls import participant_flow_urls, menu_urls, Requires
from ..utils import lighter_color as lighter_color_util, darker_color

register = template.Library()
if TYPE_CHECKING:
    from ..models import (AbstractServiceFormItem, ResponsibilityPerson, SubitemMixin, Activity,
                          ActivityChoice, ParticipationActivity, ParticipationActivityChoice,
                          Question, QuestionAnswer)


class FlowItem(NamedTuple):
    name: str
    title: str
    target: str
    attrs: Dict[str, str]


class MenuItem(NamedTuple):
    name: str
    title: str
    url: str
    is_active: bool


class MenuItems(NamedTuple):
    left: List[MenuItem]
    right: List[MenuItem]


@register.simple_tag(takes_context=True)
def responsible_link(context: Context, item: 'AbstractServiceFormItem') -> SafeString:
    """
    Used in category captions in report views, for example
    """
    responsible = context.get('responsible')
    item_responsibles = item.responsibles.all()
    links = []
    for item_responsible in item_responsibles:
        if responsible != item_responsible:
            links.append(format_html('<a class="responsible-link" href="{}">{}</a>',
                                     reverse('view_responsible', args=(item_responsible.pk,)),
                                     item_responsible))
        else:
            links.append(mark_safe(responsible))

    rv = format_html('({})', safe_join(', ', links)) if links else ''
    return rv


@register.assignment_tag
def has_responsible(item: 'SubitemMixin', responsible: 'ResponsibilityPerson') -> bool:
    return item.has_responsible(responsible)


@register.assignment_tag(takes_context=True)
def participation_items(context: Context, item: 'Union[Activity, ActivityChoice]')\
        -> 'Sequence[ParticipationActivity, ParticipationActivityChoice]':
    revision_name = utils.get_report_settings(context['request'], 'revision')
    return item.participation_items(revision_name)


@register.assignment_tag(takes_context=True)
def questionanswers(context: Context, item: 'Question') -> 'Sequence[QuestionAnswer]':
    revision_name = utils.get_report_settings(context['request'], 'revision')
    return item.questionanswers(revision_name)


@register.assignment_tag(takes_context=True)
def all_revisions(context: Context) -> bool:
    revision_name = utils.get_report_settings(context['request'], 'revision')
    return revision_name == utils.RevisionOptions.ALL


@register.assignment_tag(takes_context=True)
def participants(context: Context) -> 'Sequence[Participant]':
    revision_name = utils.get_report_settings(context['request'], 'revision')
    service_form = context.get('service_form')

    qs = Participant.objects.filter(form_revision__form=service_form,
                                    status__in=Participant.READY_STATUSES).order_by('surname')
    if revision_name == utils.RevisionOptions.CURRENT:
        qs = qs.filter(form_revision__id=service_form.current_revision_id)
    elif revision_name == utils.RevisionOptions.ALL:
        pass
    else:
        qs = qs.filter(form_revision__name=revision_name)
    return [utils.get_participant(i) for i, in qs.values_list('pk')]


@register.assignment_tag(takes_context=True)
def participant_flow_menu_items(context: Context) -> List[FlowItem]:
    current_view = context['request'].resolver_match.view_name
    participant = context['request'].participant
    cat_num = context.get('cat_num', 0)
    lst = []

    for idx, f_item in enumerate(participant_flow_urls):
        if f_item.name not in participant.flow:
            continue
        if current_view == f_item.name:
            attrs = {'current': True, 'disabled': True}
        elif not participant.can_access_view(f_item.name):
            attrs = {'greyed': True, 'disabled': True}
        else:
            attrs = {}
        if f_item.name == 'participation':
            url = reverse(f_item.name, args=(cat_num,))
        else:
            url = reverse(f_item.name)
        flv = FlowItem(f_item.name, f_item.default_args.get('title', ''), url, attrs)
        lst.append(flv)
    return lst


@register.assignment_tag(takes_context=True)
def participant_flow_categories(context: Context) -> List[FlowItem]:
    current_view = 'participation'
    service_form = context['service_form']
    cat_num = context.get('cat_num', 0)
    max_cat = context.get('max_cat', 0)
    lst = []
    for idx, category in enumerate(service_form.sub_items):
        if idx == cat_num:
            attrs = {'current': True, 'disabled': True}
        elif idx > max_cat:
            attrs = {'greyed': True, 'disabled': True}
        else:
            attrs = {}
        attrs['category'] = category
        url = reverse(current_view, args=(idx,))
        flv = FlowItem(idx, category.name, url, attrs)
        lst.append(flv)
    return lst


@register.assignment_tag(takes_context=True)
def menu_items(context: Context, menu_name: str) -> MenuItems:
    """
    This is used for:
     - Reports menu
     - Participation password_login page menu
    """
    current_view = context['request'].resolver_match.view_name
    service_form = context['service_form']
    responsible = utils.get_responsible(context['request'])

    def _check_requires(requires):
        for r in requires:
            if r == Requires.ACCESS_FULL_REPORT:
                if not (responsible and responsible.show_full_report):
                    return False
            if r == Requires.RESPONSIBLE_LOGGED_IN:
                if not responsible:
                    return False
        return True

    left = []
    right = []
    for idx, f_item in enumerate(menu_urls[menu_name]):
        name = f_item.name
        is_active = current_view == name
        arglist = f_item.default_args.get('arglist', ('slug',))
        requires = f_item.default_args.get('require', [])
        if not _check_requires(requires):
            continue

        args = tuple(getattr(service_form, i) for i in arglist)
        url = reverse(name, args=args)
        is_right = f_item.default_args.get('right')
        title = f_item.default_args.get('title', '')
        icon = f_item.default_args.get('icon')
        if icon:
            title = format_html('<span class="fa fa-{}"></span> {}', icon, title)
        itm = MenuItem(name, title, url, is_active)
        if is_right:
            right.append(itm)
        else:
            left.append(itm)
    return MenuItems(left, right)


@register.filter()
def shorten(text: str) -> SafeString:
    return format_html('<span class="tooltip-only" title="{}" >{}...</span>', text,
                       text[:5].strip())


@register.filter()
def lighter_color(cat_color: ColorStr) -> ColorStr:
    return lighter_color_util(cat_color)


@register.filter(is_safe=True)
def url_target_blank(text: str) -> str:
    return text.replace('<a ', '<a target="_blank" ')


@register.filter()
def count_color(count: int) -> SafeString:
    color_for_count = utils.color_for_count(count)
    return format_html('<span class="color-count" style="background: {};">{}</span>',
                       color_for_count, count)


@register.simple_tag()
def color_style(item: 'AbstractServiceFormItem', lighter=0, attr='background') -> SafeString:
    color = item.background_color_display
    if color:
        if lighter < 0:
            for i in range(-lighter):
                color = darker_color(color)
        else:
            for i in range(lighter):
                color = lighter_color_util(color)
        return format_html('style="{}: {};"', attr, color)
    else:
        return format_html('')


@register.filter()
def translate_bool(value: bool) -> str:
    _('True')
    _('False')
    return _(str(value))



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps

from django.conf import settings
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import resolve
from django.http import Http404, HttpResponseRedirect, HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404

from .. import models, utils


def serviceform(function=None, check_form_permission=False, init_counters=False,
                all_responsibles=True, fetch_participants=False):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(request: HttpRequest, slug: str,
                    *args, **kwargs) -> HttpResponse:
            service_form = get_object_or_404(models.ServiceForm.objects, slug=slug)
            request.service_form = service_form
            if init_counters:
                service_form.init_counters(all_responsibles)
            if fetch_participants:
                revision_name = utils.get_report_settings(request, 'revision')
                utils.fetch_participants(service_form, revision_name=revision_name)
            func_ = require_form_permissions(func) if check_form_permission else func
            return func_(request, service_form, *args)

        return wrapper

    if function:
        return actual_decorator(function)
    return actual_decorator


def require_authenticated_responsible(func):
    @wraps(func)
    def wrapper(request: HttpRequest, *args, **kwargs) -> HttpResponse:
        responsible = utils.get_responsible(request)
        if responsible:
            request.service_form = responsible.form
        if request.user.pk or responsible:
            return func(request, responsible, *args)
        else:
            raise PermissionDenied

    return wrapper


def require_authenticated_participant(function=None, check_flow=True):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(request: HttpRequest, *args, title: str='', **kwargs) -> HttpResponse:
            current_view = request.resolver_match.view_name

            participant_pk = request.session.get('authenticated_participant')
            if participant_pk:
                request.participant = participant = get_object_or_404(
                    models.Participant.objects.all(),
                    pk=participant_pk,
                    status__in=models.Participant.EDIT_STATUSES)
                if check_flow:
                    # Check flow status
                    participant._current_view = current_view
                    if not participant.can_access_view(current_view, auth=True):
                        return participant.redirect_last()
                    rv = func(request, participant, *args, **kwargs)
                    if isinstance(rv, HttpResponseRedirect):
                        url = resolve(rv.url)
                        next_view = url.view_name
                        participant.proceed_to_view(next_view)
                    return rv
                else:
                    return func(request, participant, *args, **kwargs)

            else:
                raise PermissionDenied

        return wrapper

    if function:
        return actual_decorator(function)
    return actual_decorator


def require_published_form(func):
    @wraps(func)
    def wrapper(request: HttpRequest, participant: models.Participant,
                *args, **kwargs) -> HttpResponse:
        if not participant.form.is_published:
            raise PermissionDenied
        return func(request, participant, *args, **kwargs)

    return wrapper


def require_form_permissions(func):
    @wraps(func)
    def wrapper(request: HttpRequest, service_form: models.ServiceForm,
                *args, **kwargs) -> HttpResponse:
        try:
            utils.user_has_serviceform_permission(request.user, service_form)
        except PermissionDenied:
            responsible = utils.get_responsible(request)
            if not responsible or not responsible.show_full_report:
                return redirect_to_login(request.path, login_url=settings.LOGIN_URL)

        return func(request, service_form, *args, **kwargs)

    return wrapper  #



# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

# Celery tasks

import logging

from celery import shared_task
from django.utils import timezone
from datetime import timedelta

from . import models

logger = logging.getLogger('serviceform.tasks')


@shared_task
def cleanup_abandoned_participations():
    logger.info('Deleting abandoned participations')
    models.Participant.objects.filter(last_modified__lt=timezone.now() - timedelta(days=1),
                                      status=models.Participant.STATUS_ONGOING).delete()


@shared_task
def finish_abandoned_updating_participations():
    for p in models.Participant.objects.filter(
            last_modified__lt=timezone.now() - timedelta(days=1),
            status=models.Participant.STATUS_UPDATING):
        logger.info('Finishing abandoned updating participations %s', p)
        p.finish(email_participant=False)


def test_task():
    raise Exception


# -*- coding: utf-8 -*-
# (c) 2017 Tuomas Airaksinen
#
# This file is part of Serviceform.
#
# Serviceform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Serviceform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Serviceform.  If not, see <http://www.gnu.org/licenses/>.

"""
The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.9/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from itertools import chain

from django.conf.urls import url
from django.utils.translation import ugettext_lazy as _
from django.http import HttpResponseRedirect

from django.conf import settings

from .views import login_views
from .views import participation_views
from .views import reports_views


def error(request):
    raise Exception('Test error')


def main_page(request):
    return HttpResponseRedirect(settings.LOGIN_URL)


class DummyUrl:
    """
    URL that is shown in menu but ignored in django config.
    """

    def __init__(self, name, kwargs):
        self.name = name
        self.default_args = kwargs


class Requires:
    ACCESS_FULL_REPORT = object()
    RESPONSIBLE_LOGGED_IN = object()
    EMAIL_VERIFICATION = object()


participant_flow_urls = [
    url(r'^participant/contact/$', participation_views.contact_details, name='contact_details',
        kwargs={'title': _('Contact details')}),
    url(r'^participant/email_verification/$', participation_views.email_verification,
        name='email_verification', kwargs={'title': _('Email verification')}),
    url(r'^participant/participation/(\d+)/$', participation_views.participation,
        name='participation', kwargs={'title': _('Participation details')}),
    url(r'^participant/questions/$', participation_views.questions, name='questions',
        kwargs={'title': _('Questions')}),
    url(r'^participant/preview/$', participation_views.preview, name='preview',
        kwargs={'title': _('Preview')}),
    url(r'^participant/submitted/$', participation_views.submitted, name='submitted',
        kwargs={'title': _('Ready!')}),
]

report_urls = [
    url(r'^report/([\w-]+)/$', reports_views.all_responsibles, name='report',
        kwargs={'title': _('Responsibles')}),
    url(r'^report/([\w-]+)/all_participants/$', reports_views.all_participants,
        name='all_participants', kwargs={'title': _('Participants')}),
    url(r'^report/([\w-]+)/all_activities/$', reports_views.all_activities, name='all_activities',
        kwargs={'title': _('Participations')}),
    url(r'^report/([\w-]+)/all_questions/$', reports_views.all_questions, name='all_questions',
        kwargs={'title': _('Answers')}),
    DummyUrl(name='responsible_report',
             kwargs={'title': _('My report'), 'arglist': (), 'icon': 'bullseye',
                     'require': (Requires.RESPONSIBLE_LOGGED_IN,)}),

    # Invite users
    url(r'^report/([\w-]+)/settings/$', reports_views.settings_view, name='settings',
        kwargs={'title': _('Report settings'), 'right': True, 'icon': 'cog'}),
    url(r'^invite/([\w-]+)/$', reports_views.invite, name='invite',
        kwargs={'title': _('Invite'), 'right': True, 'icon': 'user-plus'}),
    DummyUrl(name="admin:serviceform_serviceform_change",
             kwargs={'title': _('Edit form'), 'icon': 'pencil-square-o', 'right': True,
                     'arglist': ('id',)}),
    DummyUrl(name="admin:serviceform_serviceform_changelist",
             kwargs={'title': _('Admin'), 'icon': 'wrench', 'right': True, 'arglist': ()}),
    url(r'^logout/$', reports_views.logout_view, name='logout',
        kwargs={'title': _('Log out'), 'icon': 'sign-out', 'right': True, 'arglist': ()}),
]

anonymous_report_urls = [
    url(r'^for_responsible/$', reports_views.responsible_report, name='responsible_report',
        kwargs={'title': _('Your report'), 'arglist': ()}),
    url(r'^for_responsible/edit_details/$', reports_views.edit_responsible,
        name='edit_responsible',
        kwargs={'title': _('Edit your contact details'), 'icon': 'pencil-square-o',
                'arglist': ()}),
    url(r'^for_responsible/to_full_report/$', reports_views.to_full_report, name='to_full_report',
        kwargs={'title': _('To full report'), 'icon': 'bullseye', 'arglist': (),
                'require': (Requires.ACCESS_FULL_REPORT,)}),
    # DummyUrl(name='report', kwargs={'title': _('To full report'), 'arglist': ()}),
    url(r'^logout/$', reports_views.logout_view, name='logout',
        kwargs={'title': _('Log out'), 'icon': 'sign-out', 'right': True, 'arglist': ()}),
]

login_urls = [
    DummyUrl(name='password_login', kwargs={'title': _('Password login'), 'icon': 'sign-in'}),
    url(r'^([\w-]+)/send_participant_link/', login_views.send_participant_email,
        name='send_participant_email', kwargs={'title': _('Former users'), 'icon': 'key'}),
    url(r'^([\w-]+)/send_responsible_link/', login_views.send_responsible_email,
        name='send_responsible_email', kwargs={'title': _('Responsibles'), 'icon': 'user'}),
    DummyUrl(name='report',
             kwargs={'title': _('Admin login'), 'icon': 'user-secret', 'right': True}),
]

menu_urls = {'report': report_urls, 'anonymous_report': anonymous_report_urls,
             'participant_flow': participant_flow_urls, 'login': login_urls}

urlpatterns = [u for u in
               chain(participant_flow_urls, report_urls, anonymous_report_urls, login_urls) if
               not isinstance(u, DummyUrl)] + \
              [
                  # Test erorr email
                  url(r'^test_error/$', error, name='test_error'),
                  # Later actions for participant
                  url(r'^anonymous/authenticate_participant/([\w-]+)/$',
                      participation_views.authenticate_participant_old,
                      name='authenticate_participant'),
                  url(r'^anonymous/authenticate_participant/(\d+)/([\w-]+)/$',
                      participation_views.authenticate_participant,
                      name='authenticate_participant_new'),
                  url(r'^anonymous/authenticate_participant_mock/(\d+)/$',
                      participation_views.authenticate_participant_mock,
                      name='authenticate_participant_mock'),

                  # Anonymous report viewing pages for responsible persons
                  url(r'^anonymous/authenticate_responsible/([\w-]+)/$',
                      reports_views.authenticate_responsible_old, name='authenticate_responsible'),
                  url(r'^anonymous/authenticate_responsible/(\d+)/([\w-]+)/$',
                      reports_views.authenticate_responsible, name='authenticate_responsible_new'),
                  url(r'^anonymous/authenticate_responsible_mock/(\d+)/$',
                      reports_views.authenticate_responsible_mock,
                      name='authenticate_responsible_mock'),

                  url(r'^email/unsubscribe_participant/(\w+)/$', participation_views.unsubscribe,
                      name='unsubscribe_participant'),
                  url(r'^email/unsubscribe_responsible/(\w+)/$', reports_views.unsubscribe,
                      name='unsubscribe_responsible'),

                  url(r'^participant/verify_email/(\d+)/([\w-]+)/$',
                      participation_views.verify_email, name='verify_email'),

                  # Report views
                  url(r'^report/participant/(\d+)/$', reports_views.view_participant,
                      name='view_user'),
                  url(r'^report/responsible/(\d+)/$', reports_views.view_responsible,
                      name='view_responsible'),

                  # Form previews
                  url(r'^preview/([\w-]+)/$', reports_views.preview_form, name='preview_form'),
                  url(r'^preview_printable/([\w-]+)/$', reports_views.preview_printable,
                      name='preview_printable'),

                  url(r'^logout/$', reports_views.logout_view, name='logout'),
                  url(r'^send_auth_link/(.*)$', participation_views.send_auth_link,
                      name='send_auth_link'),
                  url(r'^participant/delete/$', participation_views.delete_participation,
                      name='delete_participation'),
                  url(r'^([\w-]+)/$', login_views.password_login, name='password_login'),
                  url(r'^participant/participation/$', participation_views.participation,
                      name='participation', kwargs={'cat_num': 0}),
                  url(r'^$', main_page, name='main_page'),
              ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-22 10:48
from __future__ import unicode_literals

import datetime
from django.conf import settings
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
from django.utils.timezone import utc
import select2.fields
import serviceform.serviceform.models
import serviceform.serviceform.utils



def add_basic_rights_groups_and_permissions(apps, schema_editor):
    full = ['activity',
            'activitychoice',
            'formrevision',
            'level1category',
            'level2category',
            'question',
            'responsibilityperson',
            'serviceform',
            'emailtemplate',
            ]
    delete = [
        'participant',
        'participationactivity',
        'participationactivitychoice',
        'questionanswer',
    ]

    custom_permissions = []

    Permission = apps.get_model('auth', 'Permission')
    Group = apps.get_model('auth', 'Group')
    ContentType = apps.get_model('contenttypes', 'ContentType')
    grp, created = Group.objects.get_or_create(name='Serviceform basic rights')

    def add_perm(_model, _action):
        print('Trying to add permission %s to %s' %(_action, _model))
        ct, created = ContentType.objects.get_or_create(app_label='serviceform', model=_model)
        codename = '%s_%s' % (_action, _model)
        try:
            p = Permission.objects.get(codename=codename, content_type=ct)
        except Permission.DoesNotExist:
            p = Permission.objects.create(codename=codename, name=codename.capitalize().replace('_', ' '), content_type=ct)

        grp.permissions.add(p)

    for model in full:
        for action in ['add', 'change', 'delete']:
            add_perm(model, action)

    for model in delete:
        add_perm(model, 'delete')

    ct = ContentType.objects.get(app_label='serviceform', model='serviceform')
    p = Permission.objects.create(codename='can_access_serviceform',
                                  content_type=ct, name='Can access Service Form')


def null(apps_schema_editor):
    pass



class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Activity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Name')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('order', models.PositiveIntegerField(db_index=True, default=0, verbose_name='Order')),
                ('multiple_choices_allowed', models.BooleanField(default=True, verbose_name='Multichoice')),
                ('people_needed', models.PositiveIntegerField(default=0, verbose_name='Needed')),
                ('skip_numbering', models.BooleanField(default=False, verbose_name='Skip')),
            ],
            options={
                'verbose_name': 'Activity',
                'verbose_name_plural': 'Activities',
                'ordering': ('order',),
                'abstract': False,
            },
            bases=(serviceform.serviceform.models.SubitemMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ActivityChoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Name')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('order', models.PositiveIntegerField(db_index=True, default=0, verbose_name='Order')),
                ('people_needed', models.PositiveIntegerField(default=0, verbose_name='Needed')),
                ('skip_numbering', models.BooleanField(default=False, verbose_name='Skip')),
                ('activity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Activity')),
            ],
            options={
                'verbose_name': 'Activity choice',
                'verbose_name_plural': 'Activity choices',
                'ordering': ('order',),
                'abstract': False,
            },
            bases=(serviceform.serviceform.models.SubitemMixin, models.Model),
        ),
        migrations.CreateModel(
            name='EmailMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('from_address', models.CharField(max_length=256)),
                ('to_address', models.CharField(max_length=256)),
                ('subject', models.CharField(max_length=256)),
                ('content', models.TextField()),
                ('sent_at', models.DateTimeField(null=True)),
                ('context', models.TextField(default='{}')),
            ],
        ),
        migrations.CreateModel(
            name='EmailTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Template name')),
                ('subject', models.CharField(max_length=256, verbose_name='Subject')),
                ('content', models.TextField(help_text='Following context may (depending on topic) be available for both subject and content: {{responsible}}, {{participant}}, {{last_modified}}, {{form}}, {{url}}, {{contact}}', verbose_name='Content')),
            ],
            options={
                'verbose_name': 'Email template',
                'verbose_name_plural': 'Email templates',
            },
            bases=(serviceform.serviceform.models.CopyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='FormRevision',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(max_length=32, verbose_name='Revision name')),
                ('valid_from', models.DateTimeField(default=datetime.datetime(2999, 12, 31, 22, 20, tzinfo=utc), verbose_name='Valid from')),
                ('valid_to', models.DateTimeField(default=datetime.datetime(2999, 12, 31, 22, 20, tzinfo=utc), verbose_name='Valid to')),
                ('send_bulk_email_to_participants', models.BooleanField(default=True, help_text='Send email to participants that filled the form when this revision was active. Email is sent when new current revision is published.', verbose_name='Send bulk email to participants')),
                ('send_emails_after', models.DateTimeField(default=datetime.datetime(2999, 12, 31, 22, 20, tzinfo=utc), help_text='Sends bulk email to responsibility persons at specified time, after which it will send email for each new participation', verbose_name='Email sending starts')),
            ],
            options={
                'verbose_name': 'Form revision',
                'verbose_name_plural': 'Form revisions',
                'ordering': ('-valid_from',),
            },
        ),
        migrations.CreateModel(
            name='Level1Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Name')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('order', models.PositiveIntegerField(db_index=True, default=0, verbose_name='Order')),
                ('background_color', serviceform.serviceform.models.ColorField(blank=True, null=True, verbose_name='Background color')),
            ],
            options={
                'verbose_name': 'Level 1 category',
                'verbose_name_plural': 'Level 1 categories',
                'ordering': ('order',),
                'abstract': False,
            },
            bases=(serviceform.serviceform.models.SubitemMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Level2Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Name')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('order', models.PositiveIntegerField(db_index=True, default=0, verbose_name='Order')),
                ('background_color', serviceform.serviceform.models.ColorField(blank=True, null=True, verbose_name='Background color')),
                ('category', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='serviceform.Level1Category', verbose_name='Level 1 category')),
            ],
            options={
                'verbose_name': 'Level 2 category',
                'verbose_name_plural': 'Level 2 categories',
                'ordering': ('order',),
                'abstract': False,
            },
            bases=(serviceform.serviceform.models.SubitemMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Participant',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('forenames', models.CharField(max_length=64, verbose_name='Forename(s)')),
                ('surname', models.CharField(max_length=64, verbose_name='Surname')),
                ('street_address', models.CharField(max_length=128, verbose_name='Street address')),
                ('postal_code', models.CharField(max_length=32, validators=[django.core.validators.RegexValidator(code='invalid', message='Enter a valid postal code.', regex='^\\d{5}$')], verbose_name='Zip/Postal code')),
                ('city', models.CharField(max_length=32, verbose_name='City')),
                ('email', models.EmailField(db_index=True, max_length=254, verbose_name='Email')),
                ('phone_number', models.CharField(max_length=32, validators=[django.core.validators.RegexValidator(message="Phone number must be entered in the format: '050123123' or '+35850123123'. Up to 15 digits allowed.", regex='^\\+?1?\\d{9,15}$')], verbose_name='Phone number')),
                ('auth_keys_hash_storage', django.contrib.postgres.fields.jsonb.JSONField(default=[])),
                ('secret_key', models.CharField(db_index=True, default=serviceform.serviceform.utils.generate_uuid, max_length=36, unique=True, verbose_name='Secret key')),
                ('year_of_birth', models.SmallIntegerField(blank=True, null=True, verbose_name='Year of birth')),
                ('status', models.CharField(choices=[('invited', 'invited'), ('ongoing', 'ongoing'), ('updating', 'updating'), ('finished', 'finished')], default='ongoing', max_length=16)),
                ('last_finished_view', models.CharField(default='', max_length=32)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('last_modified', models.DateTimeField(auto_now=True, verbose_name='Last modified')),
                ('last_finished', models.DateTimeField(null=True, verbose_name='Last finished')),
                ('email_verified', models.BooleanField(default=False, verbose_name='Email verified')),
                ('send_email_allowed', models.BooleanField(default=True, help_text='You will receive email that contains a link that allows later modification of the form. Also when new version of form is published, you will be notified. It is highly recommended that you keep this enabled unless you move away and do not want to participate at all any more. You can also change this setting later if you wish.', verbose_name='Sending email allowed')),
                ('form_revision', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='serviceform.FormRevision')),
            ],
            options={
                'verbose_name': 'Participant',
                'verbose_name_plural': 'Participants',
            },
        ),
        migrations.CreateModel(
            name='ParticipantLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('writer_id', models.PositiveIntegerField()),
                ('message', models.TextField()),
                ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Participant')),
                ('writer_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='ParticipationActivity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('additional_info', models.CharField(blank=True, max_length=1024, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('activity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Activity')),
                ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Participant')),
            ],
            options={
                'ordering': ('activity__category__category__order', 'activity__category__order', 'activity__order'),
            },
        ),
        migrations.CreateModel(
            name='ParticipationActivityChoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('additional_info', models.CharField(blank=True, max_length=1024, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('activity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='choices_set', to='serviceform.ParticipationActivity')),
                ('activity_choice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.ActivityChoice')),
            ],
            options={
                'ordering': ('activity_choice__order',),
            },
        ),
        migrations.CreateModel(
            name='Question',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField(db_index=True, default=0, verbose_name='Order')),
                ('question', models.CharField(max_length=1024, verbose_name='Question')),
                ('answer_type', models.CharField(choices=[('integer', 'Integer'), ('short_text', 'Short text'), ('long_text', 'Long text'), ('boolean', 'Boolean'), ('date', 'Date')], default='short_text', max_length=16, verbose_name='Answer type')),
                ('required', models.BooleanField(default=False, verbose_name='Answer required?')),
            ],
            options={
                'verbose_name': 'Question',
                'verbose_name_plural': 'Questions',
                'ordering': ('order',),
                'abstract': False,
            },
            bases=(serviceform.serviceform.models.CopyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='QuestionAnswer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('answer', models.TextField()),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Participant')),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Question')),
            ],
            options={
                'ordering': ('question__order',),
            },
        ),
        migrations.CreateModel(
            name='ResponsibilityPerson',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('forenames', models.CharField(max_length=64, verbose_name='Forename(s)')),
                ('surname', models.CharField(max_length=64, verbose_name='Surname')),
                ('street_address', models.CharField(blank=True, max_length=128, verbose_name='Street address')),
                ('postal_code', models.CharField(blank=True, max_length=32, validators=[django.core.validators.RegexValidator(code='invalid', message='Enter a valid postal code.', regex='^\\d{5}$')], verbose_name='Zip/Postal code')),
                ('city', models.CharField(blank=True, max_length=32, verbose_name='City')),
                ('email', models.EmailField(db_index=True, max_length=254, verbose_name='Email')),
                ('phone_number', models.CharField(blank=True, max_length=32, validators=[django.core.validators.RegexValidator(message="Phone number must be entered in the format: '050123123' or '+35850123123'. Up to 15 digits allowed.", regex='^\\+?1?\\d{9,15}$')], verbose_name='Phone number')),
                ('auth_keys_hash_storage', django.contrib.postgres.fields.jsonb.JSONField(default=[])),
                ('secret_key', models.CharField(db_index=True, default=serviceform.serviceform.utils.generate_uuid, max_length=36, unique=True, verbose_name='Secret key')),
                ('send_email_notifications', models.BooleanField(default=True, help_text='Send email notifications whenever new participation to administered activities is registered. Email contains also has a link that allows accessing raport of administered activities.', verbose_name='Send email notifications')),
                ('hide_contact_details', models.BooleanField(default=False, verbose_name='Hide contact details in form')),
                ('show_full_report', models.BooleanField(default=False, verbose_name='Grant access to full reports')),
            ],
            options={
                'verbose_name': 'Responsibility person',
                'verbose_name_plural': 'Responsibility persons',
                'ordering': ('surname',),
            },
            bases=(serviceform.serviceform.models.CopyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ServiceForm',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, verbose_name='Name of the serviceform')),
                ('slug', models.SlugField(help_text='Tämä on osa lomakkeen osoitetta, ts. lomakkeen osoitteeksi tulee http://localhost:8000/valitsemasi-nimi-urlissa', unique=True, verbose_name='Slug')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('last_updated', models.DateTimeField(auto_now=True, null=True, verbose_name='Last updated')),
                ('require_email_verification', models.BooleanField(default=True, verbose_name='Require email verification')),
                ('password', models.CharField(blank=True, default='', help_text='Password that is asked from participants', max_length=32, verbose_name='Password')),
                ('hide_contact_details', models.BooleanField(default=False, verbose_name='Hide contact details (other than email) in form')),
                ('flow_by_categories', models.BooleanField(default=False, help_text='Please note that preview shows full form despite this option', verbose_name='Split participation form to level 1 categories')),
                ('allow_skipping_categories', models.BooleanField(default=False, help_text='In effect only if flow by categories option is enabled. If this option is enabled, user can jump between categories. If disabled, he must proceed them one by one.', verbose_name='Allow jumping between categories')),
                ('level1_color', serviceform.serviceform.models.ColorField(blank=True, help_text='If left blank (black), default coloring will be used', null=True, verbose_name='Level 1 category default background color')),
                ('level2_color', serviceform.serviceform.models.ColorField(blank=True, help_text='If left blank (black), it will be derived from level 1 background color', null=True, verbose_name='Level 2 category default background color')),
                ('activity_color', serviceform.serviceform.models.ColorField(blank=True, help_text='If left blank (black), it will be derived from level 2 background color', null=True, verbose_name='Activity default background color')),
                ('description', models.TextField(blank=True, help_text='Description box will be shown before instruction box in participation view.', verbose_name='Description')),
                ('instructions', models.TextField(blank=True, help_text='Use HTML formatting. Leave this empty to use default. This is shown in participation view.', null=True, verbose_name='Instructions')),
                ('login_text', models.TextField(blank=True, help_text='This will be shown in the login screen', null=True, verbose_name='Login text')),
                ('required_year_of_birth', models.BooleanField(default=False, verbose_name='Year of birth')),
                ('required_street_address', models.BooleanField(default=True, verbose_name='Street address')),
                ('required_phone_number', models.BooleanField(default=True, verbose_name='Phone number')),
                ('visible_year_of_birth', models.BooleanField(default=True, verbose_name='Year of birth')),
                ('visible_street_address', models.BooleanField(default=True, verbose_name='Street address')),
                ('visible_phone_number', models.BooleanField(default=True, verbose_name='Phone number')),
                ('bulk_email_to_responsibles', models.ForeignKey(blank=True, help_text='Email that is sent to responsibles when emailing starts', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Bulk email to responsibles')),
                ('current_revision', models.ForeignKey(blank=True, help_text='You need to first add a revision to form (see below) and save. Then newly created revision will appear in the list.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='serviceform.FormRevision', verbose_name='Current revision')),
                ('email_to_former_participants', models.ForeignKey(blank=True, help_text='Email that is sent to former participants when form is published', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Bulk email to former participants')),
                ('email_to_invited_users', models.ForeignKey(blank=True, help_text='Email that is sent when user is invited to the form manually via invite form', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Invite email')),
                ('email_to_participant', models.ForeignKey(blank=True, help_text='Email that is sent to participant after he has fulfilled his participation', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Email to participant, on finish')),
                ('email_to_participant_on_update', models.ForeignKey(blank=True, help_text='Email that is sent to participant after he has updated his participation', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Email to participant, on update')),
                ('email_to_responsible_auth_link', models.ForeignKey(blank=True, help_text='Email that is sent to responsible when he requests auth link', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Responsible requests auth link')),
                ('email_to_responsibles', models.ForeignKey(blank=True, help_text='Email that is sent to responsibles when new participation is registered', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Email to responsibles')),
                ('last_editor', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='last_edited_serviceform', to=settings.AUTH_USER_MODEL, verbose_name='Last edited by')),
                ('resend_email_to_participant', models.ForeignKey(blank=True, help_text='Email that is sent to participant if he requests resending email', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Resend email to participant')),
                ('responsible', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='serviceform.ResponsibilityPerson', verbose_name='Responsible')),
                ('verification_email_to_participant', models.ForeignKey(blank=True, help_text='Email verification message that is sent to participant when filling form, if email verification is enabled', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='serviceform.EmailTemplate', verbose_name='Verification email to participant')),
            ],
            options={
                'verbose_name': 'Service form',
                'verbose_name_plural': 'Service forms',
            },
            bases=(serviceform.serviceform.models.SubitemMixin, models.Model),
        ),
        migrations.AddField(
            model_name='responsibilityperson',
            name='form',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='serviceform.ServiceForm'),
        ),
        migrations.AddField(
            model_name='question',
            name='form',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.ServiceForm'),
        ),
        migrations.AddField(
            model_name='question',
            name='responsibles',
            field=select2.fields.ManyToManyField(blank=True, related_name='question_related', sorted=False, to='serviceform.ResponsibilityPerson', verbose_name='Responsible persons'),
        ),
        migrations.AddField(
            model_name='level2category',
            name='responsibles',
            field=select2.fields.ManyToManyField(blank=True, related_name='level2category_related', sorted=False, to='serviceform.ResponsibilityPerson', verbose_name='Responsible persons'),
        ),
        migrations.AddField(
            model_name='level1category',
            name='form',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.ServiceForm'),
        ),
        migrations.AddField(
            model_name='level1category',
            name='responsibles',
            field=select2.fields.ManyToManyField(blank=True, related_name='level1category_related', sorted=False, to='serviceform.ResponsibilityPerson', verbose_name='Responsible persons'),
        ),
        migrations.AddField(
            model_name='formrevision',
            name='form',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.ServiceForm', verbose_name='Service form'),
        ),
        migrations.AddField(
            model_name='emailtemplate',
            name='form',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.ServiceForm'),
        ),
        migrations.AddField(
            model_name='emailmessage',
            name='template',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='serviceform.EmailTemplate'),
        ),
        migrations.AddField(
            model_name='activitychoice',
            name='responsibles',
            field=select2.fields.ManyToManyField(blank=True, related_name='activitychoice_related', sorted=False, to='serviceform.ResponsibilityPerson', verbose_name='Responsible persons'),
        ),
        migrations.AddField(
            model_name='activity',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='serviceform.Level2Category', verbose_name='Category'),
        ),
        migrations.AddField(
            model_name='activity',
            name='responsibles',
            field=select2.fields.ManyToManyField(blank=True, related_name='activity_related', sorted=False, to='serviceform.ResponsibilityPerson', verbose_name='Responsible persons'),
        ),
        migrations.AlterUniqueTogether(
            name='participationactivitychoice',
            unique_together=set([('activity', 'activity_choice')]),
        ),
        migrations.AlterUniqueTogether(
            name='participationactivity',
            unique_together=set([('participant', 'activity')]),
        ),
        migrations.AlterUniqueTogether(
            name='formrevision',
            unique_together=set([('form', 'name')]),
        ),
        migrations.RunPython(add_basic_rights_groups_and_permissions, null),

    ]
